#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "params.h"

/* "direction" part of the move specification (Li,Ri,Ti,Bi) */
enum direction {LEFT, RIGHT, TOP, BOTTOM, MAX_DIRECTION};

/* encoding of player colors (don't change) */
enum color {WHITE=0, BLACK=1, X=WHITE, O=BLACK, MAX_COLOR};

#define MOVE_DIR(MOVE)	((MOVE)&0x3)
#define MOVE_IDX(MOVE)	((MOVE)>>2)
#define MOVE_MAX	(BOARDSZ*4)

#define  MAX_ESTIM   INT_MAX
#define  MIN_ESTIM   INT_MIN

struct position {
    int turn;			/* color of the next mover */
    int me;			/* MY (not enemy's) color */
    int level;			/* look-ahead depth */

    int marker_count[2];	/* how much Xs and Ox are on board */

    char rowstate[BOARDSZ];
    char colstate[BOARDSZ];	/* X/O balance indicators for each row
				   and column. 0==all Xs, 2*BOARDSZ==all
				   Os, BOARDSZ==fifty-fifty (in particular,
				   empty) */

    int bits[(BOARDSZ*BOARDSZ*2+sizeof(int)*8-1)/(sizeof(int)*8)];
};
//typedef struct position *position;


/* Default constructor for struct position */

static inline void position_init(struct position *pos)
{
    int i;
    memset (pos, 0, sizeof(struct position)); //caller guarantees zeroed board */
    pos->turn=0;
    pos->me=0;

    for (i = 0; i < BOARDSZ; i++) {
	pos->rowstate[i] = pos->colstate[i] = BOARDSZ;
    }
}

/* Copy constructor for struct position */

static inline void position_copy(const struct position *old_pos,struct position *new_pos)
{
    *new_pos = *old_pos;		/* struct copy */
}

/* Check if two positions are the same */

static inline int position_match(const struct position *p1,const struct position *p2)
{
    /* good quick check first */
    if (p1->marker_count[0] != p2->marker_count[0] ||
	p1->marker_count[1] != p2->marker_count[1]) {
	return 0;
    }

    if (p1->turn != p2->turn) {
	return 0;
    }

    return (memcmp (p1->bits, p2->bits, sizeof (p1->bits)) == 0);
}

static inline int position_cell_empty (const struct position *pos, int row, int col)
{
    int bit_offset = (row*BOARDSZ + col) * 2;
    int int_offset = bit_offset / (sizeof (int) * 8);
    int shift      = bit_offset % (sizeof (int) * 8);

    return ((pos->bits[int_offset] >> shift) & 0x2) == 0;
}

static inline int position_get_marker (const struct position *pos, int row, int col)
{
    int bit_offset = (row*BOARDSZ + col) * 2;
    int int_offset = bit_offset / (sizeof (int) * 8);
    int shift      = bit_offset % (sizeof (int) * 8);

    return (pos->bits[int_offset] >> shift) & 1;
}

static inline void position_put_marker (struct position *pos, int row, int col, int color)
{
    int bit_offset = (row*BOARDSZ + col) * 2;
    int int_offset = bit_offset / (sizeof (int) * 8);
    int shift      = bit_offset % (sizeof (int) * 8);

    pos->bits[int_offset] &= ~(0x3 << shift);
    pos->bits[int_offset] |= (0x2 | color) << shift;
}


/* Do MOVE of the player pos->turn in POS */

static inline void do_move (struct position *pos, int move)
{
    int row, col;
    int color = pos->turn;
    int oldcolor = -1;	/* color of the "pushed off" marker -- none yet */

#define PUSH_MARKER \
    if (position_cell_empty (pos, row, col)) {                  \
	position_put_marker (pos, row, col, color);             \
	pos->rowstate[row] += (color<<1) - 1;                   \
	pos->colstate[col] += (color<<1) - 1;                   \
	oldcolor = -1;                                          \
	goto out;                                               \
    }                                                           \
    else {                                                      \
	oldcolor = position_get_marker (pos, row, col);         \
	position_put_marker (pos, row, col, color);             \
	pos->rowstate[row] += (color-oldcolor)<<1;              \
	pos->colstate[col] += (color-oldcolor)<<1;              \
	color = oldcolor;                                       \
    }                                                           \

    switch (MOVE_DIR (move)) {
    case LEFT:
	row = MOVE_IDX (move);
	for (col = 0; col < BOARDSZ; col++) {
	    PUSH_MARKER;
	}
	break;

    case RIGHT:
	row = MOVE_IDX (move);
	for (col = BOARDSZ-1; col >=0; col--) {
	    PUSH_MARKER;
	}
	break;

    case TOP:
	col = MOVE_IDX (move);
	for (row = 0; row < BOARDSZ; row++) {
	    PUSH_MARKER;
	}
	break;

    case BOTTOM:
	col = MOVE_IDX (move);
	for (row = BOARDSZ-1; row >= 0; row--) {
	    PUSH_MARKER;
	}
	break;
    }

out:
    if (oldcolor >= 0) {
	pos->marker_count[oldcolor]--;
    }
    pos->marker_count[pos->turn]++;
    pos->turn ^= 1;
    pos->level++;
}

static inline int estimate (struct position *pos)
{
    int estimate = 0;
    int fin = 0;
    int i;

#define SUMIT(XWINS, OWINS, ITEM) \
    switch (ITEM) {                                     \
    case 0: fin += XWINS;            break;             \
    case 1: estimate += XWINS << 24; break;             \
    case 2: estimate += XWINS << 16; break;             \
    case 3: estimate += XWINS << 8;  break;             \
    case 4: estimate += XWINS;       break;             \
    }                                                   \
    switch (2*BOARDSZ - ITEM) {                         \
    case 0: fin += OWINS;            break;             \
    case 1: estimate += OWINS << 24; break;             \
    case 2: estimate += OWINS << 16; break;             \
    case 3: estimate += OWINS << 8;  break;             \
    case 4: estimate += OWINS;       break;             \
    }                                                   \

#define DOIT(XWINS, OWINS) \
    for (i=0; i < BOARDSZ; i++) {                       \
	SUMIT (XWINS, OWINS, pos->colstate[i]);         \
	SUMIT (XWINS, OWINS, pos->rowstate[i]);         \
    }                                                   \

    switch (pos->me) {
    case 0:			/* X */
	DOIT (1, -1);
	break;
    case 1:			/* O */
	DOIT (-1, 1);
	break;
    }

    if (fin) {
	return (fin > 0? INT_MAX: INT_MIN);
    }

    return estimate;
}

void position_print2 (const struct position *pos, int indent)
{
    int row, col;

    fprintf (stderr, "%*s", indent, "");
    fprintf (stderr, "turn:%d me:%d level:%d (%d:%d)\n",
	     pos->turn, pos->me, pos->level,
	     pos->marker_count[0],
	     pos->marker_count[1]);

    fprintf (stderr, "%*s", indent, "");
    fprintf (stderr, "  ");
    for (col = 0; col < BOARDSZ; col++) {
	fprintf (stderr, "%d ", (col+1)%10);
    }
    fprintf (stderr, "\n");

    for (row = 0; row < BOARDSZ; row++) {
	fprintf (stderr, "%*s", indent, "");
	fprintf (stderr, "%d ", (row+1)%10);
	for (col = 0; col < BOARDSZ; col++) {
	    fprintf (stderr, "%c ",
		    position_cell_empty (pos, row, col) ? ' ' :
		    position_get_marker (pos, row, col) ? 'o' : 'x');
	}
	fprintf (stderr, "\n");
    }
}

void position_print (const struct position *pos)
{
    position_print2 (pos, 0);
}

