//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_block_allocator.hh,v $
// $Revision: 1.13 $
// $Date: 2002/02/04 17:29:58 $

#ifndef __gr_block_allocator_hh__
#define __gr_block_allocator_hh__

#include "pxx_i_allocator.hh"
#include "pxx_chunk_allocator.hh"
#include "pxx_spinlock.hh"
#include "pxx_sync.hh"

#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>

namespace grace
{

class BlockAllocator :
  public pxx::Spinlock,
  public pxx::I_Allocator
{

public:

  struct BlockTreeItem
  {
    union {
      BlockTreeItem* lt ;
      void* ptr ;
    };
    union {
      BlockTreeItem* rt ;
      uintptr_t ref_count ;
    };
    BlockTreeItem* up ;
    ssize_t size ;
  };

private:

  void* heap ;
  size_t heap_limit ;
  int heap_fd ;

  pxx::ChunkAllocator bti_allocator ;

  BlockTreeItem* root ;

  size_t size ;
  size_t order ;

public:

  BlockAllocator (
    void* _heap, size_t _heap_limit, int _heap_fd,
    pxx::I_Allocator& _allocator
  ) :
    heap (_heap),
    heap_limit (_heap_limit),
    heap_fd (_heap_fd),
    bti_allocator (_allocator, sizeof(BlockTreeItem)),
    root (0),
    size (pxx::page_size),
    order (pxx::get_order(size))
  {
    if (!pxx::aligned(heap, pxx::page_size))
      FATAL("heap is bad aligned");
    ftruncate(heap_fd, pxx::page_size);
    root = new_chunk(pxx::page_size);
  }

  ~BlockAllocator ()
  {
    int res;
    res = munmap(heap, size);
    if (res == -1)
      FATAL("munmap() failed, errno = %d", errno);
  }

  inline BlockTreeItem* new_chunk (
    ssize_t _size,
    BlockTreeItem* _up = null,
    BlockTreeItem* _lt = null,
    BlockTreeItem* _rt = null
  )
  {
    BlockTreeItem* res = (BlockTreeItem*)bti_allocator.allocate();
    res->size = _size;
    res->up = _up;
    res->lt = _lt;
    res->rt = _rt;
    return res;
  }

  inline void delete_chunk (BlockTreeItem* _ptr)
  {
    bti_allocator.deallocate(_ptr);
  }


  void expand ()
  {
    if ((size << 1) > heap_limit)
      FATAL("Heap exhausted");
    if (ftruncate (heap_fd, size << 1) != 0)
      FATAL("truncate() failed, errno = %d\n", errno);
    if (root->size != (ssize_t)size) {
      BlockTreeItem* p = root;
      root = new_chunk(size);
      BlockTreeItem* q = new_chunk(size, root);
      p->up = root;
      root->lt = p;
      root->rt = q;
      size <<= 1;
    } else {
      root->size = (size <<= 1);
    }
  }

  BlockTreeItem* allocate (size_t _size)
  {
    {
      synchronize(pxx::Spinlock, self);
      ssize_t s = pxx::rounds [pxx::get_order(_size)];
      ssize_t current_size = size;
      size_t offset = 0;
      BlockTreeItem* p = root;
      if (p->size >= s) {
        //
        // walk down to the tree
        while (p->size < current_size) {
          if (p->lt->size >= s) {
            p = p->lt;
            current_size >>= 1;
          } else {
            p = p->rt;
            current_size >>= 1;
            offset += current_size;
          }
        }
        BlockTreeItem* q = p;
        //
        // split a block
        while (p->size > s) {
          p->size >>= 1;
          p->lt = new_chunk(p->size, p);
          p->rt = new_chunk(p->size, p);
          p = p->lt;
        }
        p->size = -p->size;
        p->ptr = pxx::ptr_add(heap, offset);
        p->ref_count = 1;
        while (q->up != 0) {
          q = q->up;
          ssize_t sz = pxx_max(q->lt->size, q->rt->size);
          if (sz == q->size) break;
          else q->size = sz;
        }
        return p;
      }
      expand();
    }
    return allocate(_size);
  }

  void deallocate (void* _ptr)
  {
    synchronize(pxx::Spinlock, self);
    size_t offset = pxx::sub_ptr(_ptr, heap);
    BlockTreeItem* p = root;
    ssize_t current_size = size;
    while (p->size != -current_size) {
      current_size >>= 1;
      if ((offset & current_size) == 0) {
        p = p->lt;
      } else {
        p = p->rt;
      }
    }
    _deallocate(p);
  }

  void deallocate (BlockTreeItem* _p)
  {
    synchronize(pxx::Spinlock, self);
    _deallocate(_p);
  }

  void _deallocate (BlockTreeItem* _p)
  {
    if (--_p->ref_count == 0) {
      _p->size = -_p->size;
      _p = _p->up;
      while (_p != null) {
        if (_p->lt->size == _p->rt->size) {
#ifdef PARANOIA
if ((_p->size != _p->lt->size) || (_p->size != _p->rt->size))
  printf("!!!!!!!!\n");
#endif
          _p->size = (_p->size << 1);
          delete_chunk(_p->lt);
          delete_chunk(_p->rt);
          _p->lt = _p->rt = null;
          _p = _p->up;
        }
        else break;
      }
      while (_p != null) {
        ssize_t sz = pxx_max(_p->lt->size, _p->rt->size);
        if (_p->size == sz) break;
        else {
          _p->size = sz;
          _p = _p->up;
        }
      }
    }
  }

  void* reallocate (void*, size_t)
  {
    synchronize(Spinlock, self);
    return null;
  }

  void memory_dump ()
  {
    synchronize(Spinlock, self);
    _memory_dump(root);
    printf("\n\n");
  }

  void _memory_dump (BlockTreeItem* _ptr)
  {
    if (_ptr != null) {
      if (_ptr->lt == null) {
        if (_ptr->size == 0) printf("a:%d ", -_ptr->size);
        else printf("f:%d ", _ptr->size);
      } else {
        _memory_dump(_ptr->lt);
        _memory_dump(_ptr->rt);
      }
    } else {
      printf("Empty");
    }
  }

};

}

#endif // __gr_block_allocator_hh__

