#include "gr_comp_resource.hh"
#include "gr_message.ih"
#include "gr_continuation.ih"
#include "gr_computation_state.ih"
#include "gr_node.ih"
#include "gr_options.hh"
#include <climits>

namespace grace {

//! Get index for any given ComputationResource
unsigned ComputationResource::get_index() {
  int n = this - comp_res_data->resources;
  if (n < 0 || n >= (2*MAX_NODE-1))
    n = get_src();
  return (unsigned)n;
}

ComputationResource * ComputationResource::root() {
  return comp_res_data->resources + 2*(MAX_NODE-1);
}

ComputationResource * ComputationResource::up() {
  return comp_res_data->resources + (MAX_NODE | (get_index() >> 1));
}

ComputationResource * ComputationResource::left() {
  return comp_res_data->resources + ((~MAX_NODE & get_index()) << 1);
}

ComputationResource * ComputationResource::right() {
  return left() + 1;
}

bool ComputationResource::leafp() {
  int n = this - comp_res_data->resources;
  if (n < 0 || n >= (2*MAX_NODE-1))
    FATAL("leafp: out of tree");
  return get_index() < MAX_NODE;
}

void ComputationResource::merge() {
  free_mflops = pxx_max(left()->free_mflops, right()->free_mflops);
  dmesg("merge: %d is now %d mflops\n", get_index(), free_mflops);
}

void ComputationResource::recalc() {
//  dmesg("Recalc resource %d\n", get_index());
  if (!leafp() && !valid) {
    left()->recalc();
    right()->recalc();
    merge();
    valid = true;
  }
//  dmesg("Recalc completed\n");
}

int ComputationResource::weight(ResourceRequest*) {
  return free_mflops;
}

ComputationResource::ComputationResource() : 
  Message(sizeof(ComputationResource), GR_MSG_AUTODELETE),
  free_mflops(INT_MIN), valid(false)
{}

void ComputationResource::invalidate() {
  ComputationResource *t = up();  
  while(true) {
    int n = t->get_index();
    if (n < 0 || n >= (2*MAX_NODE-1)) 
      FATAL("invalidate: out of tree");
    t->valid = false;
    if (t == root())
      break;
    t = t->up();
  }   
}

void ComputationResource::handle() {
  {
    synchronize(pxx::Spinlock,comp_res_data->comp_res_lock);
//if (comm_layer->get_id() == comm_layer->get_main_id())
dmesg("Received resource update info from CS %u\n", get_src());
    ComputationResource *cr = comp_res_data->resources + get_index();
    cr->free_mflops = free_mflops;
    cr->valid       = true;
    cr->invalidate();
    dmesg("Resource tree update completed\n");
  }
//  if (free_mflops > 0) {
    ContinuationBase *cont;
    if ((cont = prenatal_queue->get_front()) != null) {
      if (!cont->get_node_external()) {
        prenatal_queue->put_front(cont);
      } else {
dmesg("ComputationResource::handle: external computation (type=%d)\n",cont->get_type());
      }
    }
//  }
}

ComputationResource *
ComputationResource::find_best_rec(ResourceRequest *rq, unsigned _depth) {
//  dmesg("find_best_free for %d\n", get_index());
  if (leafp())
    return this;
  ComputationResource *l = left(), *r = right();
//  return l->weight(rq) >= r->weight(rq) ?
//    l->find_best_rec(rq) : r->find_best_rec(rq);
  if ((l->weight(rq) > r->weight(rq)) ||
//      ((l->weight(rq) == r->weight(rq)) && ((1 << _depth) ) & (comm_layer->get_id())))
      ((l->weight(rq) == r->weight(rq)) && (1 & rand())))
  {
    return l->find_best_rec(rq, _depth + 1);
  } else {
    return r->find_best_rec(rq, _depth + 1);
  }
}

ComputationResource *
ComputationResource::find_best_free(ResourceRequest *rq) {
  synchronize(pxx::Spinlock,comp_res_data->comp_res_lock);    
  root()->recalc();
  unsigned depth = 0;
  ComputationResource *cr = root()->find_best_rec(rq, depth);
  return cr;
}

void ComputationResource::indent(int n) {
  while(n--) printf("\t");
}

void ComputationResource::show_res_tree(int level) {
  indent(level);
//  printf("src=#0x%x, free_mflops=%d\n", 
//	 get_index(), free_mflops);
  if (leafp())
    return;
  left()->show_res_tree(level+1);
  right()->show_res_tree(level+1);
}

void ComputationResource::notify_one_worker_busy(int busy) {
  synchronize(pxx::Spinlock,comp_res_data->comp_res_lock);
  comp_res_data->local_free_mflops -= CPU_MFLOPS * busy; 
//  if (comp_res_data->local_free_mflops < 0)
//    FATAL("Negative free_mflops detected\n");
#if 1
  ComputationResource *cr = new ComputationResource;
#else
  ComputationResource *cr = comp_res_data->resources + comm_layer->get_id();
#endif
  cr->free_mflops = comp_res_data->local_free_mflops; 
dmesg("Available mflops changed to %d\n", comp_res_data->local_free_mflops);
  comm_layer->broadcast(cr);
}

  
//! Returns cluster node number either -1 
//! Should be called after prenatal is taken from queue
int ComputationResource::
allocate_node_for_external_computing(CallInfo const& _info) {
  if (_info.flags & GR_CI_LOCAL) {
    return -1;
  } else if (_info.flags & GR_CI_FIXED) {
    if (_info.id != comm_layer->get_id()) {
      return (int)(_info.id);
    } else {
      return -1;
    }
//  } else if (_info.flags & GR_CI_SYSTEM) {
//    return (int)(comm_layer->get_id());
  }
  ComputationResource::root()->recalc();
  if (ComputationResource::root()->free_mflops <= 0) {
    return -1; 
  }
  ComputationResource *cr = 
    ComputationResource::find_best_free(NULL); 
  if (!cr) {
    return -1;
  }
  if (cr->get_index() == comm_layer->get_id()) {
    return -1;
  }
  if (cr->free_mflops <= 0) {
    return -1;
  }

#if 0
  /* FIXME: lines below is a temporary hack */
  if (cr->get_index() <= comm_layer->get_id())
    return -1;
#endif

#if 0
  if (!prenatal_queue->is_empty() ||
      cr->free_mflops > comp_res_data->local_free_mflops)
#else
  unsigned n = prenatal_queue->get_size() +
               calculator->get_size() +
               local_queue->get_size();
  if (
    (n >= options.high_wm) ||
    ((n >= options.low_wm) && (cr->free_mflops > comp_res_data->local_free_mflops))
//    (n >= GR_LOAD_LOW) || (cr->free_mflops > comp_res_data->local_free_mflops)
  )
#endif
  {
    //
    // FIXME: do not schedule to continuation originator
    if (_info.src_id == cr->get_index()) {
dmesg("Recheduling continuation to its originator %u\n", _info.src_id);
//      return -1;
    }
    /* Speculative decrement */
    cr->free_mflops -= CPU_MFLOPS;
    cr->invalidate();

    return cr->get_index();
  }
  return -1;
}

}
