//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_continuation.hh,v $
// $Revision: 1.37 $
// $Date: 2002/01/29 17:05:59 $

#ifndef __gr_continuation_hh__
#define __gr_continuation_hh__

#include "gr_common.hh"
#include "gr_shared_allocator.hh"
#include "gr_message.hh"
#include "gr_computation_state.hh"
#include "gr_call_info.hh"
#include "gr_profile.hh"

#include <stdlib.h>

namespace grace
{

//
//! Forward reference to node class
class Node ;

//! A common part of continuation

class ContinuationBase :
  public Message, public pxx::Spinlock
{

//
//! T-value is our friend
template <typename val_t> friend class T_Val ;
template <typename val_t> friend class T_Ptr ;

protected:
  //
  //! Pointer to a node which computes a value
  Node* node ;
#ifdef CHECK_INVARIANTS
public:
#endif // CHECK_INVARIANTS
  //
  //! Reference counter
  volatile unsigned ref_count ;

protected:
  //
  //! Pointer to a factory which will create a computation state
  GenericComputationStateFactory* factory ;

public:

  //
  //! tcall information
  CallInfo info ;

  //
  //! Pointer to a function
  void (*function) (void const*, void*) ;

  //
  //! Default constructor
  inline ContinuationBase (size_t _size) ;
  //
  //! Constructor
  inline ContinuationBase (
    GenericComputationStateFactory* _factory,
    size_t _size,
    CallInfo const& _info,
    void (*_function)(void const*, void*)
  ) ;
  //
  //! Destructor
  inline ~ContinuationBase () ;
  //
  //! Does continuation has an attached node
  inline bool has_node () volatile ;
  //
  //! Getting attached node
  inline Node* get_node () ;
  //
  //! Getting attached node (w/o lock)
  inline Node* get_node_internal () ;
  //
  //! Try to get an external node
  inline bool get_node_external () ;
  //
  //! Get computation type
  inline ComputationType get_type () const ;
  //
  //! Set computation state factory
  inline void set_factory (GenericComputationStateFactory* _factory) ;
  //
  //! Create computation state
  inline GenericComputationState*  create_cstate (Node* _node) ;
  //
  //! Do external computation
  inline bool do_external_computation (CompSpaceID _dst) ;
  //
  //! Try to force external computation at fixed node
  inline void try_fixed () ;
  //
  //!
  virtual void const* get_args () const = 0 ;
  //
  //!
  virtual void* get_ress () = 0 ;
  //
  //! Overloaded operator new
  OVERLOAD_NEW
  //
  //! Overloaded operator delete
  OVERLOAD_DELETE

//private:

  //
  //! Increase reference counter
  inline void add_ref () volatile ;
  //
  //! Decrease reference counter
  inline bool del_ref () volatile ;

#ifdef CHECK_INVARIANTS
  unsigned bind_count ;
#endif // CHECK_INVARIANTS

};

//
//! Continuation template

template <typename arg_struct_t, typename res_struct_t>
struct Continuation :
  public ContinuationBase
{
  //
  //!
  arg_struct_t args ;
  //
  //!
  res_struct_t ress ;
  //
  //! Default constructor
  Continuation () ;
  //
  //! constructor
  Continuation (
    void (*_function) (arg_struct_t const*, res_struct_t*),
    GenericComputationStateFactory* _factory,
    CallInfo const& _info
  ) ;
  //
  //! constructor
  Continuation (
    void (*_function) (arg_struct_t const*, res_struct_t*),
    GenericComputationStateFactory* _factory
  ) ;
  //
  //! Destructor
  ~Continuation () ;

  //
  //! A method which serializes an object to a message
  Message* serialize () ;
  //
  //! A method which deserializes a message to a real object
  Message* deserialize () ;
  //
  //! Handle remote request
  void handle() ;
  //
  //!
  void const* get_args () const ;
  //
  //!
  void* get_ress () ;

};

}

#endif // __gr_continuation_hh__

