#include "gr_core.hh"
#include "gr_signal.hh"
#include "gr_worker.hh"
#include "gr_node.ih"
#include "gr_continuation.ih"
#include "gr_computation_state.ih"
#include "gr_main_process.ih"
#include "gr_system_process.ih"
#include "gr_external_process.ih"
#include "gr_control_message.hh"
#include "gr_comp_resource.hh"
#include "gr_message.ih"
#include "gr_memo.hh"
#include "gr_tptr.hh"
#include "gr_timer.hh"
#include "gr_options.hh"
#include "gr_common.hh"
#include "gr_profile.hh"

#include <sys/time.h>
#include <sched.h>
#include <pthread.h>
#include <limits.h>
#include <pwd.h>

//#include <mpi.h>

#define USE_PTHREAD

namespace grace

{

SharedAllocator* shared_allocator ;

char tmsg[1024];
BlockAllocator* block_allocator ;
SharedChunkAllocator* ring_node_allocator ;
SharedChunkAllocator* continuation_allocator ;
SharedChunkAllocator* node_allocator ;
#ifdef MEM_DEBUG
MemCounter* continuation_mem_count ;
MemCounter* node_mem_count ;
#endif // MEM_DEBUG
unsigned cpu_number ;
Worker* workers ;
pid_t* comm_pid ;
pthread_t system_tid;
pid_t* system_pid ;
pid_t* main_pid = null;
SharedPipe* sys_proc_pipe ;
SharedPipe* main_pipe ;
SharedPipe* workers_stop_pipe ;
atomic_t*   workers_to_stop ;
pxx::Spinlock* global_lock ;
pxx::Spinlock* dmesg_lock ;
pxx::Spinlock current_lock ;
void* comm_handle ;
CommLayer* comm_layer ;

size_t heap_size ;
size_t stack_size ;
Node* current = null ;
Worker* current_worker ;
ComputationResourceData *comp_res_data ;

pxx::Ring<ContinuationBase>* prenatal_queue ;
pxx::Ring<ContinuationBase>* upload_queue ;
pxx::Ring<ContinuationBase>* local_queue ;
pxx::Ring<Node>* calculator ;
pxx::Ring<Node>* upload_calculator ;
pxx::Ring<Node>* collector ;

Node* main_node ;

static void (**atexit_func)(void) = null ;

static void get_options(int& argc, char**& argv) ;

static char const* make_file_name (
  char const* const _dir, char const* const _file
)
{
  unsigned ld = strlen(_dir);
  unsigned lf = strlen(_file);
  unsigned len = ld +lf + 2;
  if (len > PATH_MAX) FATAL("File name length exceeded");
  char const* p = new char[len];
  if (p == null) FATAL("new[]() failed");
  memcpy((void*)p, _dir, ld);
  ((char*)p)[ld] = '/';
  memcpy((void*)(p + ld + 1), _file, lf);
  ((char*)p)[len - 1] = 0;
  return p;
}

Core::Core (int& _argc, char**& _argv) :
  argc (_argc),
  argv (_argv),
  shared_allocator ((SharedAllocator*)0x48000000)
{
  vmesg("Initializing GRACE core...\n");
  {
    vmesg("- Parsing GRACE command line options...");
    get_options(argc, argv);
    vmesg("done\n");
  }
  {
    vmesg("- Creating program working directory... ");
    work_dir = new char[PATH_MAX + 1];
    int dangerous_size = PATH_MAX - 64;
    char* p = work_dir;
    static char const* const common_dir = "/tmp";
    unsigned n = strlen(common_dir);
    if ((dangerous_size -= n) < 0) FATAL("Buffer overflow detected");
    memcpy(p, common_dir, n);
    p += n;
    *p++ = '/';
    char* s = strrchr(argv[0], '/');
    if (s == null) s = argv[0];
    else s += 1;
    n = strlen(s);
    if ((dangerous_size -= n) < 0) FATAL("Buffer overflow detected");
    memcpy(p, s, n);
    p += n;
    *p++ = '-';
    struct passwd* pwd = getpwuid(getuid());
    if (pwd->pw_name != null) {
      n = strlen(pwd->pw_name);
      if ((dangerous_size -= n) < 0) FATAL("Buffer overflow detected");
      memcpy(p, pwd->pw_name, n);
    }
    else n = sprintf(p, "%u", pwd->pw_uid);
    p += n;
    *p++ = '-';
    p += sprintf(p, "%u", getpid());
    *p++ = '-';
    *p++ = '-';
    time_t t = time(null);
    struct tm* tm = localtime(&t);
    p += sprintf(
      p, "%04u-%02u-%02u--%02u-%02u-%02u",
      tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
      tm->tm_hour, tm->tm_min, tm->tm_sec
    );
    *p = 0;
    int res = mkdir(work_dir, 0700);
    if (res != 0) FATAL("mkdir() failed, errno = %d\n", errno);
    vmesg("done\n");
  }
  {
    shared_allocator_buf_filename =
      make_file_name(work_dir, "shared_allocator_buf.mem");
    shared_heap_filename =
      make_file_name(work_dir, "shared_heap.mem");
    block_heap_filename =
      make_file_name(work_dir, "block_heap.mem");
  }
  {
    vmesg("- Creating shared heap allocator... ");
    shared_allocator_buf_fd = open(
      shared_allocator_buf_filename,
      O_CREAT | O_TRUNC | O_RDWR, S_IRUSR | S_IWUSR
    );
    if (shared_allocator_buf_fd == -1)
      FATAL("open() failed, errno = %d\n", errno);
    if (ftruncate(shared_allocator_buf_fd, pxx::page_size) != 0)
      FATAL("ftruncate() failed, errno = %d", errno);
    void* res = mmap(
      shared_allocator, pxx::page_size,
      PROT_READ | PROT_WRITE, MAP_SHARED | MAP_FIXED, shared_allocator_buf_fd, 0
    );
    if (res == MAP_FAILED)
      FATAL("mmap() failed, errno = %d\n", errno);
    shared_heap_fd = open(
      shared_heap_filename,
      O_CREAT | O_TRUNC | O_RDWR, S_IRUSR | S_IWUSR
    );
    if (shared_heap_fd == -1)
      FATAL("open() failed, errno = %d\n", errno);
    res = mmap(
      shared_heap_start, shared_heap_size,
      PROT_READ | PROT_WRITE, MAP_SHARED | MAP_FIXED, shared_heap_fd, 0
    );
    if (res == MAP_FAILED)
      FATAL("mmap() failed, errno = %d\n", errno);
    grace::shared_allocator = new(shared_allocator) SharedAllocator(
      shared_heap_start, shared_heap_size, shared_heap_fd
    );
    vmesg("done\n");
  }
  {
    vmesg("- Creating block allocator... ");
    block_heap_fd = open(
      block_heap_filename,
      O_CREAT | O_TRUNC | O_RDWR, S_IRUSR | S_IWUSR
    );
    if (block_heap_fd == -1)
      FATAL("open() failed, errno = %d\n", errno);
    void* res = mmap(
      block_heap_start, block_heap_size,
      PROT_READ | PROT_WRITE, MAP_SHARED | MAP_FIXED,
      block_heap_fd, 0
    );
    if (res == MAP_FAILED)
      FATAL("mmap() failed, errno = %d\n", errno);
    block_allocator =
      new(shared_allocator->allocate(sizeof(BlockAllocator)))
        BlockAllocator(
          block_heap_start, block_heap_size, block_heap_fd,
          *shared_allocator
        );
    vmesg("done\n");
  }
  {
    vmesg("- Unlinking mmaped files... ");
    if (!options.persistent) {
      if (unlink(shared_allocator_buf_filename) != 0)
        FATAL("unlink() failed, errno = %d\n", errno);
      if (unlink(shared_heap_filename) != 0)
        FATAL("unlink() failed, errno = %d\n", errno);
      if (unlink(block_heap_filename) != 0)
        FATAL("unlink() failed, errno = %d\n", errno);
    }
    vmesg("done\n");
  }
  {
    vmesg("- Creating ring node allocator... ");
    ring_node_allocator =
      new(shared_allocator->allocate(sizeof(SharedChunkAllocator)))
        SharedChunkAllocator(
          *shared_allocator, sizeof(pxx::Ring<int>)
        );
    vmesg("done\n");
  }
  {
    vmesg("- Creating continuation allocator... ");
    continuation_allocator =
      new(shared_allocator->allocate(sizeof(SharedChunkAllocator)))
        SharedChunkAllocator(
          *shared_allocator, sizeof(ContinuationBase)
        );
    vmesg("done\n");
  }
  {
    vmesg("- Creating node allocator... ");
    node_allocator =
      new(shared_allocator->allocate(sizeof(SharedChunkAllocator)))
        SharedChunkAllocator(
          *shared_allocator, sizeof(Node)
        );
    vmesg("done\n");
  }
#ifdef MEM_DEBUG
  {
    continuation_mem_count = new MemCounter("Continuation");
    node_mem_count = new MemCounter("Node");
  }
#endif // MEM_DEBUG
  {
    vmesg("- Creating main queues...\n");
    prenatal_queue =
      new(shared_allocator->allocate(sizeof(pxx::Ring<ContinuationBase>)))
        pxx::Ring<ContinuationBase>(*ring_node_allocator);
    vmesg("- - Prenatal queue created\n");
    upload_queue =
      new(shared_allocator->allocate(sizeof(pxx::Ring<ContinuationBase>)))
        pxx::Ring<ContinuationBase>(*ring_node_allocator);
    local_queue =
      new(shared_allocator->allocate(sizeof(pxx::Ring<ContinuationBase>)))
        pxx::Ring<ContinuationBase>(*ring_node_allocator);
    calculator =
      new(shared_allocator->allocate(sizeof(pxx::Ring<Node>)))
        pxx::Ring<Node>(*ring_node_allocator);
    vmesg("- - Calculator created\n");
    upload_calculator =
      new(shared_allocator->allocate(sizeof(pxx::Ring<Node>)))
        pxx::Ring<Node>(*ring_node_allocator);
    collector =
      new(shared_allocator->allocate(sizeof(pxx::Ring<Node>)))
        pxx::Ring<Node>(*ring_node_allocator);
    vmesg("- - Collector created\n");
    vmesg("- Queues created\n");
  }
  {
    if (options.nworkers != 0) {
      cpu_number = options.nworkers;
    } else {
      vmesg("- Checking for CPUs number... ");
      int fd = open("/proc/cpuinfo", O_RDONLY);
      if (fd == -1) FATAL("open() failed");
      char buf[65536];
      memset(buf, sizeof(buf), 0);
      read(fd, buf, sizeof(buf) - 1);
      cpu_number = 0;
      char* s = buf;
#if defined(__i386__)
      while ((s = strstr(s, "processor")) != null) s++, cpu_number++;
#elif defined(__alpha__)
      s = strstr(s, "active");
      if (s != null) sscanf(s, "active %u", &cpu_number);
#endif
      close(fd);
      vmesg("%u\n", cpu_number);
    }
  }
  {
    vmesg("- Creating pseudo node for main process... ");
    main_node = new Node;
    vmesg("done\n");
  }
  {
    vmesg("- Creating external process factory... ");
    external_factory = new ComputationStateFactory<ExternalProcess>;
    vmesg("done\n");
  }
  {
    vmesg("- Creating system process factory... ");
    system_factory = new ComputationStateFactory<SystemProcess>;
    vmesg("done\n");
  }
  {
    vmesg("- Creating computation resources data... ");
    comp_res_data = new ComputationResourceData;
    vmesg("done\n");
  }
  {
    size_t sz = MEMO_TABLE_SIZE * sizeof(MemoTableEntryBase*);
    MemoTableEntryBase::table =
      (MemoTableEntryBase**)shared_allocator->allocate(sz);
    memset(MemoTableEntryBase::table, 0, sz);
    MemoTableEntryBase::memo_spinlock =
      (pxx::Spinlock*)shared_allocator->allocate(sizeof(pxx::Spinlock));
    new (MemoTableEntryBase::memo_spinlock) pxx::Spinlock;
    tptr_unique_id_count =
      (atomic_t*)shared_allocator->allocate(sizeof(atomic_t));
    atomic_t tmp = { 0 };
    *tptr_unique_id_count = tmp;
  }
  {
    comm_pid = new(shared_allocator->allocate(sizeof(int))) pid_t;
    system_pid = new(shared_allocator->allocate(sizeof(int))) pid_t;
    main_pid = new(shared_allocator->allocate(sizeof(int))) pid_t;
    atexit_func = (void (**)())shared_allocator->allocate(sizeof(void*));
    workers_to_stop = new(shared_allocator->allocate(sizeof(atomic_t))) 
      atomic_t;
    *atexit_func = null;
    *system_pid = 0;
    *main_pid = 0;
    atomic_set(workers_to_stop, 0);
    sys_proc_pipe = new SharedPipe;
    main_pipe = new SharedPipe;
    workers_stop_pipe = new SharedPipe;
  }
  vmesg("Initialization completed\n");
}

Core::~Core ()
{
//  if (getpid() == *main_pid) {
//    vmesg("Signalling to communication layer\n");
//    comm_layer->signal();
//    return;
//  }
  delete main_node;
#ifdef MEM_DEBUG
  delete continuation_mem_count;
  delete node_mem_count;
#endif // MEM_DEBUG
  if (*main_pid != 0) {
    vmesg("Killing main process, pid = %u\n", *main_pid);
    kill (*main_pid, SIGINT);
  }
  {
    unsigned i;
    for (i = 0; i < cpu_number; i++) {
      vmesg("Killing worker process, pid = %u\n", workers[i].get_pid());
      kill (workers[i].get_pid(), SIGINT);
    }
  }
  if (system_pid != null) {
    ContinuationBase* cont = null;
    TRY(sys_proc_pipe->write(&cont, sizeof(ContinuationBase*)));
#ifdef USE_PTHREAD
    pthread_join(system_tid, null);
#endif
  }
  gr_comm_layer_destroy();
  shared_allocator->~SharedAllocator();
  if (!options.persistent) {
    int res = rmdir(work_dir);
    if (res != 0) FATAL("rmdir() failed, errno=%d", errno);
  }
  delete shared_allocator_buf_filename;
  delete shared_heap_filename;
  delete block_heap_filename;
}

extern "C" pid_t __libc_fork();
extern "C" int __libc_pause(void);

void Core::run ()
{
  vmesg("Starting worker processes...\n");
  global_lock =
    (pxx::Spinlock*)shared_allocator->allocate(sizeof(pxx::Spinlock));
  dmesg_lock =
    (pxx::Spinlock*)shared_allocator->allocate(sizeof(pxx::Spinlock));
  new (global_lock) pxx::Spinlock();
  new (dmesg_lock) pxx::Spinlock();
  workers =
    (Worker*)shared_allocator->allocate(sizeof(Worker) * cpu_number);
  for (unsigned i = 0; i < cpu_number; i++) {
    global_lock->lock();
    pid_t res = __libc_fork();
dmesg("001\n");
    if (res == -1)
      FATAL("fork() failed, errno = %d\n", errno);
    if (res == 0) {
dmesg("002\n");
      new(workers + i) Worker(getpid());
dmesg("003\n");
      workers[i].run();
dmesg("004\n");
    }
    global_lock->lock();
    global_lock->unlock();
  }
  pxx::terminate = &terminate;
  vmesg("Worker processes started\n");
}

void terminate (int _error_code)
{
  pid_t cur_pid = getpid();
  for (unsigned i = 0; i < cpu_number; i++) {
    if (workers[i].get_pid() != cur_pid) {
      vmesg("Killing worker process, pid = %u\n", workers[i].get_pid());
      kill (workers[i].get_pid(), SIGTERM);
    }
  }
  if (*comm_pid != cur_pid) {
    vmesg("Killing communication process, pid = %u\n", *comm_pid);
    kill (*comm_pid, SIGTERM);
  }
  if (*main_pid != cur_pid) {
    vmesg("Killing main process, pid = %u\n", *main_pid);
    kill (*main_pid, SIGTERM);
  }
  if (*system_pid != cur_pid) {
    vmesg("Killing system process, pid = %u\n", *system_pid);
    kill (*system_pid, SIGKILL);
  }
  vmesg("Exiting from current process, pid = %u\n", cur_pid);
  exit(_error_code);
}

void abort ()
{
  int res;
  ControlMessage <GR_MSG_ABORT> *msg = new ControlMessage <GR_MSG_ABORT>;
  comm_layer->broadcast(msg);
  pid_t cur_pid = getpid();
  for (unsigned i = 0; i < cpu_number; i++) {
    if (workers[i].get_pid() != cur_pid) {
      vmesg("Killing worker process, pid = %u\n", workers[i].get_pid());
      res = kill (workers[i].get_pid(), SIGKILL);
      vmesg("res = %d, errno = %d\n", res, errno);
      vmesg("Killed\n");
    }
  }
  if ((*main_pid) != cur_pid && (*main_pid != 0)) {
    vmesg("Killing main process, pid = %u\n", *main_pid);
    res = kill (*main_pid, SIGKILL);
    vmesg("res = %d, errno = %d\n", res, errno);
  }
  if ((*system_pid) != cur_pid && (*system_pid != 0)) {
    vmesg("Killing system process, pid = %u\n", *system_pid);
    res = kill (*system_pid, SIGKILL);
    vmesg("res = %d, errno = %d\n", res, errno);
  }
#if 0
  if (*comm_pid != cur_pid) {
    vmesg("Killing communication process, pid = %u\n", *comm_pid);
    kill (*comm_pid, SIGKILL);
  }
#endif
  vmesg("Killing current process, pid = %u\n", cur_pid);
  res = kill (cur_pid, SIGKILL);
  vmesg("res = %d, errno = %d\n", res, errno);
//  ::abort();
}

void finish ()
{
  global_lock->trylock();
  ControlMessage <GR_MSG_EXIT> *msg = new ControlMessage <GR_MSG_EXIT>;
  comm_layer->broadcast(msg);
  global_lock->lock();
}


int atexit (void (*_f)(void))
{
  *atexit_func = _f;
  return 0;
}

static void get_options(int& argc, char**& argv)
{
  int shift = 0;
  for (int i = 1; i < argc; i++) {
    if (shift > 0) argv[i] = argv[i + shift];
    if (strncmp(argv[i], "-grace-", 7) == 0 ||
        strncmp(argv[i], "+grace-", 7) == 0)
    {
      argv[i][0] = '-';
      if (strcmp(argv[i], "-grace-workers") == 0) {
        shift++; argc--;
        argv[i] = argv[i + shift];
        sscanf(argv[i], "%u", &options.nworkers);
      } else if (strcmp(argv[i], "-grace-low-wm") == 0) {
        shift++; argc--;
        argv[i] = argv[i + shift];
        sscanf(argv[i], "%u", &options.low_wm);
      } else if (strcmp(argv[i], "-grace-high-wm") == 0) {
        shift++; argc--;
        argv[i] = argv[i + shift];
        sscanf(argv[i], "%u", &options.high_wm);
      } else if (strcmp(argv[i], "-grace-no-notify-empty") == 0) {
        options.notify_empty = false;
      } else if (strcmp(argv[i], "-grace-persistent") == 0) {
        options.persistent = true;
      } else if (strcmp(argv[i], "-grace-verbose") == 0) {
        options.verbose = true;
      } else if (strcmp(argv[i], "-grace-show-profiling") == 0) {
        options.show_profiling = true;
      } else if (strcmp(argv[i], "-grace-profile-verbose") == 0) {
        options.profile_verbose = true;
      } else if (strcmp(argv[i], "-grace-check-stack") == 0) {
        options.check_stack = true;
      } else if (strcmp(argv[i], "-grace-stack-size") == 0) {
        shift++; argc--;
        argv[i] = argv[i + shift];
        sscanf(argv[i], "%u", &options.stack_size);
      } else {
        FATAL("Unknown GRACE option");
      }
      shift++; i--; argc--;
    }
  }
  if (options.check_stack && options.stack_size == 1)
    options.stack_size = 2;
}

}

using namespace grace ;
extern int tmain (int argc, char* argv[]) ;

jmp_buf jbuf2;

void tmain_trampoline (int argc, char* argv[]) {
dmesg("argc = %d\n", argc);
  tmain(argc, argv);
  if (options.show_profiling) {
    emesg("Main %d:\n", getpid());
    ProfBlock::print_prof_info();
  }
  __libc_longjmp(main_node->get_jmp_buf(), 0);
}

struct Args
{
  int argc ;
  char** argv ;
};

void tmain_start (volatile Args* _args)
{
  //
  //FIXME: This is a workaround to resolve problems with latest SCI-MPICH
  {
    int i;
    for (i = 1; i < _args->argc; i++) {
      if (strcmp(_args->argv[i], "--") == 0) {
        _args->argc -= i;
        _args->argv[i] = _args->argv[0];
        _args->argv += i;
        break;
      }
    }
  }
vmesg("main_pid = %d\n", getpid());
  *main_pid = getpid();
  char c;
  main_pipe->read(&c, sizeof(char));

  jmp_buf& jbuf = main_node->get_jmp_buf();
  jbuf2->PEIP = (long)tmain_trampoline;
  ComputationState<MainProcess>* cs =
    (ComputationState<MainProcess>*)(main_node->get_cstate());
  long* stk_top = (long*)(cs->get_stk());
#if defined(__i386__)
  *(--stk_top) = (long)(_args->argv);
  *(--stk_top) = (long)(_args->argc);
  *(--stk_top) = 0;
  jbuf2->PEIP = (long)tmain_trampoline;
  jbuf2->PESP = (long)stk_top;
  jbuf2->PEBP = 0;
#elif defined(__alpha__)
#endif

  //ret_val = tmain(argc, argv);
#if defined(__i386__)
  if (setjmp(jbuf) == 0) __libc_longjmp(jbuf2, 0);
#elif defined(__alpha__)
//  fprintf(stderr,"--- %p\n", _args->argv);
  if (setjmp(jbuf) == 0) {
    __asm__ __volatile__(
      "ldl $16,%0\n"
      "ldq $17,%1\n"
      "ldq $30,%2\n"
      "jsr $26,%3\n"
      :
      : "m"(_args->argc), "m"(_args->argv), "m"(stk_top), "i"(tmain_trampoline)
    );
  }
#endif

  //comm_layer->signal();
  ControlMessage <GR_MSG_EXIT> *msg = new ControlMessage <GR_MSG_EXIT>;
  comm_layer->broadcast(msg);
  __libc_pause();
//  _exit(0);
}

void sys_proc_start ()
{
  *system_pid = getpid();
  vmesg("System process started, pid = %d\n", *system_pid);
  while (true) {
    ContinuationBase* cont;
    try {
      sys_proc_pipe->read(&cont, sizeof(ContinuationBase*));
dmesg("System continuation popped, addr = %p\n", cont);
    } catch (pxx::SysError const& err) {
        FATAL("System error code = %d", err.get_code());
    }
#ifdef USE_PTHREAD
    if (cont == null) pthread_exit(null);
#endif
    {
//      synchronize(pxx::Spinlock, current_lock);
      current = cont->get_node();
//      ComputationResource::notify_one_worker_busy(true);
      current->compute();
//      ComputationResource::notify_one_worker_busy(false);
    }
//    if (cont->get_src() == comm_layer->get_id()) cont->del_ref();
  }
}

int main (int argc, char* argv[])
{
  //
  // FIXME: exit code after global t-program termination must be exactly
  // a return value from tmain
  Core core(argc, argv);
  //
  // create communication layer
  gr_comm_layer_create();
  //
  // initialize signal handlers; this should be done after
  // communication layer initialization because communication layer
  // may do some magic with signal delivering
  set_signals();
  //
  // run worker processes
  core.run();
  {
    static BlockAllocator::BlockTreeItem* stk =
      block_allocator->allocate(16 * pxx::page_size);
    if (options.check_stack) {
      int res = mprotect(
        (char*)(stk->ptr), pxx::page_size, PROT_NONE
      );
      if (res == -1) FATAL("mprotect() failed, errno = %d\n", errno);
    }
    static Args args = { argc, argv };
    int res = clone(
      (int (*)(void *))&tmain_start, (char*)(stk->ptr) + (-stk->size),
      CLONE_FS|CLONE_FILES, &args
    );
    if (res == -1) FATAL("clone() failed, errno = %d\n", errno);
  }
#ifdef USE_PTHREAD
  pthread_create(&system_tid, NULL, (void* (*)(void*))sys_proc_start, NULL);
#endif
  comm_layer->init(argc, argv);
  //
  // here is some magic with forking main process which leads
  // to some code duplication; if somebody tells how this
  // can be done in more graceful way I'll correct a code
  Timer* elapsed = null;
#if 1
  if (comm_layer->get_id() == comm_layer->get_main_id()) {
    elapsed = new Timer("Time elapsed");
vmesg("master, pid = %d\n", getpid());
    // FIXME: is it good?
    char c;
    main_pipe->write(&c, sizeof(char));
  }
#endif
vmesg("comm_pid = %d\n", getpid());
  *comm_pid = getpid();
  gr_comm_layer_loop();
  if (*atexit_func != null) {
    fprintf(stderr, "Calling atexit function\n");
    (**atexit_func)();
  }
  if (comm_layer->get_id() == comm_layer->get_main_id()) {
    delete elapsed;
  }
  fprintf(stderr, "CompSpace %u: %u nodes computed\n",
    comm_layer->get_id(), comp_res_data->nodes_computed);
  return 0;
}
