//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_memo.hh,v $
// $Revision: 1.17 $
// $Date: 2002/02/04 17:29:58 $

#ifndef __gr_memo_hh__
#define __gr_memo_hh__

#include "gr_core.hh"
#include "pxx_sync.hh"

namespace grace
{

#define MEMO_TABLE_SIZE 65537

class MemoTableEntryBase {

protected:

  //! Bucket linkage field
  MemoTableEntryBase *next ;

  //! Key data 
  void   *key ;
  size_t size ;

  //
  //! Reference counter
  unsigned ref_count ;

  //! Compute hash value 
  static unsigned hash(void *key, size_t size) ;

  MemoTableEntryBase(void* _key, size_t _size) : 
    next(null), key(_key), size(_size), ref_count (0) {}

public:

  //! Get pointer to hash table entry  
  static MemoTableEntryBase **lookup(void *key, size_t size) ;

  static MemoTableEntryBase** table;

  static pxx::Spinlock* memo_spinlock;

  //! We redefined our memory allocation to make it clean
  OVERLOAD_NEW
  OVERLOAD_DELETE

  inline void add_ref ()
  {
    ref_count++;
  }

  inline void del_ref ()
  {
    ref_count--;
  }

};


//! Intermediate structure : full memo key
template <typename arg_t, typename res_t> 
struct MemoClosure {
  arg_t arg ;
  void (*fun)(arg_t const&,res_t&);
  MemoClosure (arg_t _arg, void (*_fun)(arg_t const&,res_t&)) {
    /* The purpose of this memset is to zero any unused bytes in the
       object, because they matter when the closure is used as a key.
       There *will* be unused space after `fun' because GCC pads
       structures up to the alignment boundary */
     memset(this,0,sizeof(*this));
     arg = _arg;
     fun = _fun;
  }
  MemoClosure (MemoClosure<arg_t,res_t> const& closure) {
     arg = closure.arg;
     fun = closure.fun;
  }
};


template <typename arg_t, typename res_t>
class MemoTableEntry : public MemoTableEntryBase {

public:

  MemoClosure<arg_t,res_t> closure ;
  res_t res ;

  MemoTableEntry<arg_t,res_t>(MemoClosure<arg_t,res_t> const &_closure) :
    MemoTableEntryBase(&closure,sizeof(closure)),
    closure (_closure)
  {}

  static MemoTableEntry<arg_t,res_t>& get_entry(MemoClosure<arg_t,res_t>&);

  static inline res_t& get_value(MemoClosure<arg_t,res_t>& _closure) {
    return get_entry(_closure).res;
  }

};


#define DEFINE_MEMO_FUNCTION(_res_t,_fun,_arg_t) \
   _res_t &memo_##_fun(_arg_t _arg) {\
     MemoClosure<_arg_t,_res_t> closure(_arg,_fun); \
     return MemoTableEntry<_arg_t,_res_t>::get_value(closure); \
   }

#define GET_MEMO_ENTRY(_res_t,_fun,_arg_t,_arg) ({ \
     MemoClosure<_arg_t,_res_t> closure(_arg,_fun); \
     MemoTableEntry<_arg_t,_res_t>::get_entry(closure); \
})

template <typename args_t, typename ress_t>
class MemoMessage : public Message {

public:
  MemoClosure<args_t,ress_t> closure;
  ress_t ress;
  
  //
  //! Constructor.
  MemoMessage (args_t const& _args, void (*_fun)(args_t const&,ress_t&), ress_t& _ress) 
    : Message(sizeof(self)), closure(_args,_fun)
  {
    ress = _ress;
  }
  //
  //! Serialization. Currently does nothing.
  Message* serialize () {}
  //
  //! Deserialization. Currently does nothing.
  Message* deserialize () {}
  //
  //! Assign incoming value.
  void handle() {
    /* Lock a table and intern memo info to memotable */
    synchronize(pxx::Spinlock, *MemoTableEntryBase::memo_spinlock);
    MemoTableEntry<args_t,ress_t> **p = (MemoTableEntry<args_t,ress_t>**)
      MemoTableEntryBase::lookup(&closure,sizeof(closure));
    if (*p == null) {
      *p = new MemoTableEntry<args_t,ress_t>(closure);
      (*p)->res = ress;
      dmesg("%s: new memo info inserted to memotable\n");
    } else {
      dmesg("%s: memo info already present at memotable\n");
    }
  }
  
  //FIXME: delete something ?

};


}

#endif // __gr_memo_hh__
