#include "gr_mpi_comm_layer.ih"
#include "gr_continuation.ih"
#include "gr_node.ih"
#include "gr_core.hh"

namespace grace
{

char MPI_CommLayer::sys_buf[4096] ;

//
//! Constructor
MPI_CommLayer::MPI_CommLayer () :
  sig_count (0),
  in_queue (*ring_node_allocator),
  out_queue (*ring_node_allocator),
  initialized (false)
{
  //
  // create a signalling pipe
  TRY(sig_pipe = new SharedPipe);
}

//
//! Initialization function
void MPI_CommLayer::init (int& _argc, char**& _argv)
{
  //
  // do MPI initialization
  vmesg("Initializing MPI... ");
  MPI_CHECK(MPI_Init(&_argc, &_argv));
  MPI_CHECK(MPI_Barrier(MPI_COMM_WORLD));
  //
  // ask MPI for number of nodes and current node ID
  int sz;
  MPI_CHECK(MPI_Comm_size(MPI_COMM_WORLD, &sz));
  size = sz;
  MPI_CHECK(MPI_Comm_rank(MPI_COMM_WORLD, &rank));
  vmesg("done\n");
  initialized = true;
}

//
//! Initialization check
bool MPI_CommLayer::is_initialized () const
{
  return initialized;
}

//
//! Destructor
MPI_CommLayer::~MPI_CommLayer ()
{
  //
  // do MPI finalization
  MPI_CHECK(MPI_Barrier(MPI_COMM_WORLD));
  vmesg("%u: Finalizing MPI... \n", getpid());
  int res = MPI_Finalize();
  if (res != 0)
    FATAL("MPI_Finalize() failed, error = %d", res);
  //
  // the following line is commented out due to some broken MPI
  // implementations (e.g. LAM) - they do not allow printing after
  // MPI has been finalized
//  vmesg("done\n");
  //
  // destroy a signalling pipe
  delete sig_pipe;
}

//
//! Signal a communication layer to close all communications and exit a loop
void MPI_CommLayer::finish () const
{
  //
  // write into signalling pipe a special command
  sig_lock.lock();
  char c = 'x';
  sig_count++;
  sig_lock.unlock();
  sig_pipe->write(&c, 1);
}

//
//! Receive incoming messages
void MPI_CommLayer::recv (int _tag) const
{
//dmethod;
  int flag;
  MPI_Status status;
  Message* msg;
  while (true) {
    //
    // probe for a MPI message
//dmesg("before Iprobe\n");
    MPI_CHECK(MPI_Iprobe(
      MPI_ANY_SOURCE, _tag, MPI_COMM_WORLD, &flag, &status
    ));
//dmesg("after Iprobe\n");
    //
    // if message pending
    if (flag != 0) {
//dmesg("Iprobe success\n");
      int tag = status.MPI_TAG;
      //
      // get message size
      int count;
      MPI_CHECK(MPI_Get_count(&status, MPI_CHAR, &count));
//dmesg("after Get_count\n");
      if (tag == GR_MSG_SYSTEM) {
        //
        // set its proper size
        msg = new(sys_buf) Message(count);
dmesg("%u: Got a system message from %d\n", comm_layer->get_id(), msg->get_src());
      } else if (tag == GR_MSG_CONTINUATION) {
        //
        // allocate message object
        msg = (Message*)(shared_allocator->allocate(count));
        new(msg) Message(count);
        COUNT_NEW(continuation);
      } else {
        //
        // allocate message object
        msg = (Message*)(shared_allocator->allocate(count));
        //
        // set its proper size
        new(msg) Message(count);
      }
      //
      // and finally receive a message
      MPI_CHECK(MPI_Recv(
        msg, count, MPI_CHAR, status.MPI_SOURCE, tag, MPI_COMM_WORLD, &status
      ));
      //
      // deserialize a message in place
      msg->deserialize();
dmesg("%u: Got a message from %d\n", comm_layer->get_id(), msg->get_src());
      dmesg("%u: Got a message\n", comm_layer->get_id());
      //
      // and handle it
      msg->handle();
      if (tag == GR_MSG_AUTODELETE) delete msg;
    } else {
//dmesg("Iprobe failed\n");
      break;
    }
  }
//dmesg("recv_exit\n");
}

//
//! A loop for message processing (may be empty)
//! \todo MPI_CHAR -> MPI_BYTE
//! \todo Probably we can simply reschedule without a pipe magic. Is it good?
void MPI_CommLayer::loop () const
{
  char c;
  while (true) {
    //
    // try to read a control character from a signalling pipe
    TRY(sig_pipe->read(&c, 1));
    sig_lock.lock();
    sig_count--;
    sig_lock.unlock();
    //
    // in a special case we should exit from a loop
    if (c == 'x') break;
    //
    // otherwise...
    else {
      Message* msg;
//      recv(GR_MSG_AUTODELETE);
//      recv(MPI_ANY_TAG);
      //
      // first try to send outstanding messages
      while ((msg = out_queue.get()) != null) {
        recv(GR_MSG_AUTODELETE);
        recv(MPI_ANY_TAG);
        //
        // set a proper source for a message
        msg->set_src(comm_layer->get_id());
	//
	// check for local messages
	if (msg->get_dst() == comm_layer->get_id()) {
//vmesg("%u: Got a message from %d\n", comm_layer->get_id(), msg->get_src());
dmesg("%u: Got a local message\n", comm_layer->get_id());
	  //
          // and handle it
	  msg->handle();
	  continue;
	}
        //
        // serialize a message in place
	msg->serialize();
        //
        // if this is a unicast message - send it
	if (msg->get_dst() != MPI_BROADCAST_ID) {
          int len = msg->get_size();
dmesg("Sending message to %x, len = %d\n", msg->get_dst(), len);
          MPI_CHECK(MPI_Send(
            msg, len, MPI_CHAR, msg->get_dst(),
            msg->tag, MPI_COMM_WORLD
          ));
	} else {
	  unsigned i;
          unsigned len = msg->get_size();
	  for (i = 0; i < get_size(); i++) {
dmesg("Bcasting message to %x, len = %d\n", i, len);
            MPI_CHECK(MPI_Send(
              msg, len, MPI_CHAR, i, msg->tag, MPI_COMM_WORLD
            ));
	  }
	}
        //
        // deserialize a message in place
	msg->deserialize();
        if (msg->tag == GR_MSG_AUTODELETE) {
          delete msg;
        } else if (msg->tag == GR_MSG_CONTINUATION) {
          ContinuationBase* cont = (ContinuationBase*)msg;
          if (!cont->has_node()) {
            delete cont;
          }
        }
      }
      //
      // then receive incoming messages
      recv(GR_MSG_AUTODELETE);
      recv(MPI_ANY_TAG);
    }
  }
//  void* p = null;
//  sys_proc_pipe->write(&p, sizeof(void*));
  MPI_CHECK(MPI_Barrier(MPI_COMM_WORLD));
  global_lock->trylock();
  global_lock->unlock();
  vmesg("Exiting CommLayer::loop()\n");
}

//
//! Static method creating specific communication layer
void gr_comm_layer_create ()
{
  comm_layer = new MPI_CommLayer();
}

//
//! Static method destroying specific communication layer
void gr_comm_layer_destroy ()
{
  delete comm_layer;
}

//
//! Static method running communication loop
void gr_comm_layer_loop ()
{
  comm_layer->loop();
}

}
