//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_profile.hh,v $
// $Revision: 1.5 $
// $Date: 2002/01/25 12:37:29 $

#ifndef __gr_profile_hh__
#define __gr_profile_hh__

#include "gr_continuation.hh"
#include "gr_shared_chunk_allocator.hh"
#include "gr_options.hh"
#include "pxx_ring.hh"

#include <sys/time.h>
#include <unistd.h>

extern "C" {
  int __gettimeofday(struct timeval *, struct timezone *);
}

namespace grace
{

// Useful macros for user & compiler

#define GR_PROFILE(name)                          \
   static ProfBlock __prof_block(name,__LINE__);  \
   ProfEnterLeave   __enter_leave(__prof_block)

#define GR_PROFILE_FUNCTION GR_PROFILE(__FUNCTION__)
#define GR_PROFILE_STATEMENT(s) ({ __prof_block.set_line(__LINE__); s })

// Useful macro for tkernel

#define TPROF_CURRENT (ProfBlock::get_current_location())

class ProfTime : public timeval {

public:  

  inline ProfTime() 
  {
    tv_sec = tv_usec = 0;
  }

  inline void set_current_time() {
    __gettimeofday(this, null);
  }

  inline ProfTime& operator-= (ProfTime &pt) 
  {
    tv_sec  -= pt.tv_sec;
    tv_usec -= pt.tv_usec;
    if (tv_usec < 0) {
      tv_usec += 1000000;
      tv_sec  -= 1;
    }
    return self;
  }

  inline ProfTime& operator+= (ProfTime &pt)
  {
    tv_sec  += pt.tv_sec;
    tv_usec += pt.tv_usec;
    if (tv_usec >= 1000000) {
      tv_usec -= 1000000;
      tv_sec  += 1;
    }
    return self;
  }

};

class ProfBlock;
extern volatile ProfBlock *prof_blocks;

class ProfBlock : public ProfTime {

friend class ProfEnterLeave ;

private:

  const char *name;
  int   line;
  volatile bool in_ring;
  uint32_t count;
public: // FIXME

  unsigned long sent_bytes;
  unsigned long sent_dirty_bytes;
  unsigned long sent_count;
  unsigned long claimed_bytes;
  unsigned long claimed_dirty_bytes;
  unsigned long claimed_count;

  static ProfTime start_time;

  volatile ProfBlock *next;

public:

  ProfBlock(const char *_name, int _line) :
    name(_name), count(0)
  {
    if (!in_ring) next = null;
    set_line(_line);
  }

  inline void set_line(int _line)
  {
    line = _line;
    if (options.profile_verbose) {
      static char buf[1024];
      int len = sprint_line(buf);
      buf[len++] = '\n'; buf[len++] = 0;
      emesg(buf);
    }
  }

  static ProfBlock*& get_current_location();

  inline void start()
  {
    if (!in_ring) {
      next = prof_blocks;
      prof_blocks = this;
      in_ring = true;
    }
    start_time.set_current_time();
  }

  inline void stop()
  {
    ProfTime time;
    time.set_current_time();
    time -= start_time;
    self += time;
  }

  int sprint_line(char *_buf)
  {
    return tsprintf(_buf, "%s:%d", name, line);
  }

  static void print_prof_info()
  {
    static char buf[256];
    volatile ProfBlock* p = prof_blocks;
    while (p) {
      if (p->name) {
        tsprintf(
          buf, "[%3u:%-5u]  %-20s %10u %6u.%06u\n"
               "  %lu %lu(%lu) upload  %lu %lu(%lu) download\n",
          comm_layer->get_id(), getpid(),
          p->name, p->count, p->tv_sec, p->tv_usec,
          p->sent_count, p->sent_bytes, p->sent_dirty_bytes,
          p->claimed_count, p->claimed_bytes, p->claimed_dirty_bytes
        );
        tmesg("%s", buf);
      }
      p = p->next;
    }
  }

};

class ProfEnterLeave {

public:

  ProfBlock *parent_block ;
  bool same ;

  ProfEnterLeave(ProfBlock& _block)
  {
    if (&_block != TPROF_CURRENT) {
      same = false;
      if (TPROF_CURRENT) TPROF_CURRENT->stop();
      parent_block = TPROF_CURRENT; 
      TPROF_CURRENT = &_block;
      TPROF_CURRENT->start();
    } else {
      same = true;
    }
    TPROF_CURRENT->count++;
  }

  ~ProfEnterLeave() 
  {
    if (!same) {
      TPROF_CURRENT->stop();
      TPROF_CURRENT = parent_block;
      if (TPROF_CURRENT) TPROF_CURRENT->start();
    }
  }

};

}

#endif // __gr_profile_hh__ 



