#include "gr_signal.ih"
#include "gr_core.hh"

#include <errno.h>
#include <signal.h>
#include <sys/ucontext.h>

namespace grace
{

#if __GLIBC_MINOR__ >= 2
#define __sigaction __libc_sigaction
extern "C" int __sigaction (
  int __sig, __const struct sigaction *__restrict __act,
  struct sigaction *__restrict __oact
) __THROW;
#endif

char sig_stk[2 * SIGSTKSZ] ;
size_t const sig_stk_size = sizeof(sig_stk) ;
stack_t sig_stk_struct = { sig_stk, 0, sizeof(sig_stk) } ;
struct sigaction sigsegv_action ;
struct sigaction sigbus_action ;
struct sigaction sigpipe_action ;

static void print_prof_state()
{
  if (TPROF_CURRENT) {
    static char buf[1024];
    int len = tsprintf(buf,"signal state: ");
    len += TPROF_CURRENT->sprint_line(buf+len);
    buf[len++] = '\n';
    emesg(buf);
  }
}

void sigsegv_handler (int, siginfo_t* _siginfo, void* _uc)
{
  mcontext_t* sc = &(((struct ucontext*)_uc)->uc_mcontext);
  emesg("SIGSEGV occured, addr = %p\n", _siginfo->si_addr);
  if (_siginfo->si_addr >= block_heap_start &&
      _siginfo->si_addr < (char*)block_heap_start + block_heap_size)
  {
    emesg("*** GRACE error: Possible stack overflow detected\n");
  }
#if defined(__i386__)
  emesg("ip = 0x%x, dx = 0x%x, bx = 0x%x, ax = 0x%x, pid = %d\n", sc->gregs[14],sc->gregs[9], sc->gregs[8], sc->gregs[11], getpid());
#elif defined(__alpha__)
  emesg("ip = 0x%lx\n", sc->sc_pc);
#endif
  emesg("tmsg = %s\n", tmsg);
  abort();
}

void sigpipe_handler (int, siginfo_t* _siginfo, void* _uc)
{
  mcontext_t* sc = &(((struct ucontext*)_uc)->uc_mcontext);
  emesg("SIGPIPE occured, addr = %p\n", _siginfo->si_addr);
#if defined(__i386__)
  emesg("ip = 0x%x, bx = 0x%x, ax = 0x%x\n", sc->gregs[14], sc->gregs[8], sc->gregs[11]);
#elif defined(__alpha__)
  emesg("ip = 0x%lx\n", sc->sc_pc);
#endif
  emesg("tmsg = %s\n", tmsg);
  abort();
}

void sigbus_handler (int, siginfo_t* _siginfo, void* _uc)
{
  mcontext_t* sc = &(((struct ucontext*)_uc)->uc_mcontext);
  emesg("SIGBUS occured, addr = %p\n", _siginfo->si_addr);
#if defined(__i386__)
  emesg("ip = 0x%x, bx = 0x%x, ax = 0x%x\n", sc->gregs[14], sc->gregs[8], sc->gregs[11]);
#elif defined(__alpha__)
  emesg("ip = 0x%lx\n", sc->sc_pc);
#endif

  //  extern void attach_local_gdb();
  //  attach_local_gdb();

  abort();
}

void set_signals ()
{
  vmesg("Initializing signal handling...\n");
  //
  // set alternative stack for signal handling
  vmesg("- Setting signal stack... ");
  int res = sigaltstack(&sig_stk_struct, null);
  if (res != 0) FATAL("sigaltstack() failed, errno = %d\n", errno);
  vmesg("done\n");
  //
  // set signal handler for SIGSEGV
  {
    vmesg("- Setting signal handler for SIGSEGV... ");
    sigsegv_action.sa_handler = (void (*)(int))sigsegv_handler;
#ifndef __alpha__
    sigsegv_action.sa_restorer = null;
#endif
    sigfillset(&sigsegv_action.sa_mask);
    sigsegv_action.sa_flags = SA_RESTART | SA_SIGINFO | SA_ONSTACK;
    sigsegv_action.sa_sigaction = sigsegv_handler;
//    sigset_t set;
    res = __sigaction(SIGSEGV, &sigsegv_action, null);
    if (res != 0) FATAL("sigaction() failed, errno = %d\n", errno);
//    sigfillset (&set);
//    sigemptyset (&set);
//    sigaddset (&set, SIGSEGV);
//    sigprocmask (SIG_UNBLOCK , &set, null);
    vmesg("done\n");
  }
  // set signal handler for SIGBUS
  {
    vmesg("- Setting signal handler for SIGBUS... ");
    sigbus_action.sa_handler = (void (*)(int))sigbus_handler;
#ifndef __alpha__
    sigsegv_action.sa_restorer = null;
#endif
    sigfillset(&sigbus_action.sa_mask);
    sigbus_action.sa_flags = SA_RESTART | SA_SIGINFO | SA_ONSTACK;
    sigbus_action.sa_sigaction = sigbus_handler;
//    sigset_t set;
    res = __sigaction(SIGBUS, &sigbus_action, null);
    if (res != 0) FATAL("sigaction() failed, errno = %d\n", errno);
//    sigfillset (&set);
//    sigemptyset (&set);
//    sigaddset (&set, SIGBUS);
//    sigprocmask (SIG_UNBLOCK , &set, null);
    vmesg("done\n");
  }
  // set signal handler for SIGPIPE
 {
    vmesg("- Setting signal handler for SIGPIPE... ");
    sigpipe_action.sa_handler = (void (*)(int))sigpipe_handler;
#ifndef __alpha__
    sigsegv_action.sa_restorer = null;
#endif
    sigfillset(&sigpipe_action.sa_mask);
    sigpipe_action.sa_flags = SA_RESTART | SA_SIGINFO | SA_ONSTACK;
    sigpipe_action.sa_sigaction = sigpipe_handler;
//    sigset_t set;
    res = __sigaction(SIGPIPE, &sigpipe_action, null);
    if (res != 0) FATAL("sigaction() failed, errno = %d\n", errno);
//    sigfillset (&set);
//    sigemptyset (&set);
//    sigaddset (&set, SIGBUS);
//    sigprocmask (SIG_UNBLOCK , &set, null);
    vmesg("done\n");
  }
  vmesg("Signal handling initialized\n");
}

}
