//
// Copyright (C) 2000, 2001 GRACE development team  <grace@msu.ru>
//
// GRACE is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GRACE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GRACE; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/gr_worker.cc,v $
// $Revision: 1.9 $
// $Date: 2002/01/25 12:37:30 $

#include "gr_worker.hh"
#include "gr_computation_state.hh"
#include "gr_core.hh"
#include "gr_message.hh"
#include "gr_comp_resource.hh"
#include "gr_node.ih"
#include "gr_continuation.ih"
#include "gr_profile.hh"
#include "pxx_sys_error.ih"
#include "pxx_pipe.ih"
#include <asm/atomic.h>

namespace grace
{

void print_stat(int) 
{
  if (options.show_profiling) {
    emesg("Worker %d:\n", getpid());
    ProfBlock::print_prof_info();
  }
  kill(getpid(), SIGKILL);
  exit(0);
}

Worker::Worker (pid_t _pid) :
  pid (_pid), busy(1)
{
  current_worker = this;
  dmesg("Worker at %p, pid = %u is ready\n", this, pid);
}

Worker::~Worker ()
{}

pid_t Worker::get_pid ()
{
  return pid;
}


inline bool Worker::set_busy (int new_busy) {
  if (busy != new_busy) {
    busy = new_busy;
    ComputationResource::notify_one_worker_busy(busy);
    return true;
  } else {
    return false;
  }
}

extern atomic_t*   workers_to_stop ;
extern SharedPipe* workers_stop_pipe ;

/*!
 * This functions is called as worker loop
 * iterator to get permission for it's work. 
 * It also blocks worker when it is required.
 */
inline bool Worker::is_work_enabled ()
{
  if (atomic_read(workers_to_stop) > 0) {
    atomic_dec(workers_to_stop);
    unsigned char c;
    tmesg("Worker %d stopped\n", getpid());
    workers_stop_pipe->read(&c,1);
    tmesg("Worker %d continued\n", getpid());
  }
  return true;
}

/*! 
 * This function stops `n' workers
 */
void Worker::stop_workers (int n) 
{
  while (n-- > 0) atomic_inc(workers_to_stop);
}

/*!
 * This functions continues `n' workers
 */
void Worker::cont_workers (int n) 
{
  unsigned char c;
  while (n-- > 0) workers_stop_pipe->write(&c,1);
}


/*!
 * This function loops trying to get a node ready to compute or
 * a continuation in a prenatal state.
 * \todo Add a collector queue processing
 * \todo Replace queues implementation by priority queues
 */
void Worker::run ()
{
  vmesg("- Worker process ready, pid = %u\n", pid);
  global_lock->unlock();
  //
  // Set termination handler
  signal(SIGINT,print_stat);
  //
  // main worker process loop
  while (!comm_layer->is_initialized());
  while (is_work_enabled()) {
    {
      //
      // poll for incoming messages
      comm_layer->poll();
      if ((current = upload_calculator->get_front()) == null) {
        ContinuationBase* cont = upload_queue->get_front();
        if (cont != null) {
#ifdef DEBUG_LAST_INFO
tsprintf(tmsg,"%s:%d %p", __FUNCTION__, __LINE__, cont->function);
#endif
          if (cont->del_ref()) {
#ifdef CHECK_INVARIANTS
tmesg("!!! Worker::run(): Upload reference counter = 0\n");
#endif // CHECK_INVARIANTS
            continue;
          }
#ifdef CHECK_INVARIANTS
          if (cont->ref_count != cont->bind_count) {
            tmesg("!!! Worker::run()/upload/: ref_count != bind_count\n");
          }
#endif // CHECK_INVARIANTS
          current = cont->get_node();
        } else {
          //
          // try to get node from calculator...
          if ((current = calculator->get_front()) == null) {
            //
            // ...if there is no any, try to get a prenatal continuation...
            cont = local_queue->get_front();
            if (cont == null) {
#ifdef DEBUG_LAST_INFO
tsprintf(tmsg,"%s:%d %p", __FUNCTION__, __LINE__, null);
#endif
              cont = prenatal_queue->get_front();
            }
#ifdef DEBUG_LAST_INFO
tsprintf(tmsg,"%s:%d %p", __FUNCTION__, __LINE__, cont ? cont->function : null);
#endif
            //
            // ...and get a node for it
            if (cont != null) {
#ifdef DEBUG_LAST_INFO
tsprintf(tmsg,"%s:%d %p", __FUNCTION__, __LINE__, cont->function);
#endif
              if (cont->del_ref()) continue;
#ifdef CHECK_INVARIANTS
              if (cont->ref_count != cont->bind_count) {
                tmesg("!!! Worker::run()/usual/: ref_count != bind_count\n");
              }
#endif // CHECK_INVARIANTS
              current = cont->get_node();
              if (cont->get_type() == SystemProcess) {
                TRY(sys_proc_pipe->write(&cont, sizeof(ContinuationBase*)));
dmesg("System continuation pushed, addr = %p\n", cont);
                continue;
              }
            }
          } else {
dmesg("worker: got node from calculator = %p\n", current);
          }
        }
      }
      //
      // now finally check if we got a node
      if (current != null) {
        unsigned n = calculator->get_size() +
                     prenatal_queue->get_size() +
                     local_queue->get_size() +
                     1;
        if (n >= options.high_wm) {
          set_busy(1);
        } else if (n < options.low_wm) {
          if (!set_busy(-1)) {
            if (options.notify_empty)
              ComputationResource::notify_one_worker_busy(0);
          }
        }
        dmesg("Worker %u: got node at %p\n", pid, current);
        //
        // prepare setjmp() buffer...
        if (setjmp(jbuf) == 0) {
          //
          // ...and compute a node
          current->compute();
        } else {
          //
          // here we are returned from a node computation
          // first look if we have already finished computation...
          if (current->is_computed()) {
//            current->delete_cstate();
            if (current != main_node) delete current;
          //
          // if a node is not computed, then it is waiting for a value
//          } else {
//            dmesg("%u: Node %x is waiting...\n", pid, current);
          }
        }
      } else {
        set_busy(-1);
      }
    }
  }
}

}
