//
// Copyright (C) 2000, 2001 Andrey Slepuhin <pooh@msu.ru>
//
// libp++ is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// libp++ is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libp++; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/pxx_heap_allocator.cc,v $
// $Revision: 1.18 $
// $Date: 2001/07/16 13:27:33 $
// Author: Andrey Slepuhin <pooh@msu.ru>

#include "pxx_heap_allocator.hh"

#include <new>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>

namespace pxx
{

HeapAllocator::HeapAllocator (
  void* const _start,
  size_t const _heap_size,
  int _fd = 0,
  bool _need_remap = true
) :
  Heap (_start, _heap_size, 0, _fd, _need_remap),
  min_free_size (1 << get_order(sizeof(BlockInfo)))
{
  BlockInfo* p = reinterpret_cast<BlockInfo*>(start);
  (lists + get_order(page_size))->insert(p, page_size);
}

HeapAllocator::~HeapAllocator ()
{}

void* HeapAllocator::allocate (size_t _size)
{
  //
  // first try to allocate a block within current heap
  {
    size_t order, i;
    //
    // advance allocated size by number of bytes reserved for length
    _size += sizeof(uintptr_t);
    //
    // we cannot allocate less than minimal free size
    if (_size < min_free_size) _size = min_free_size ;
    //
    // choose first appropriate block list
    order = get_order(_size);
    _size = rounds[order];
    i = order;
    //
    // walk through apropriate block lists
    while (i <= current_order) {
      FreeBlockList* p = lists + i;
      BlockInfo* q = p->next;
      //
      // ok, we found non-empty free block list
      if (q != p) {
        //
        // remove its first element
        q->remove();
        unsigned s = rounds[i];
        while (s > _size) {
          s >>= 1;
          (--p)->insert(ptr_add(q, s), s);
        }
        //
        // set a size of allocated block
        q->set_size(_size);
        //
        // and return shifted pointer
        return user_ptr(q);
      }
      i++;
    }
  }
  //
  // if current heap exhausted try to expand it
  {
    //
    // here we did not find any non-empty free block list,
    // so we should expand a heap
    size_t old_size = current_size;
    size_t old_order = current_order;
    expand();
    BlockInfo* p = static_cast<BlockInfo*>(start);
    if (p->size == old_size) {
      p->remove();
      (lists + current_order)->insert(p, current_size);
    } else {
      (lists + old_order)->insert(ptr_add(p, old_size), old_size);
    }
    return HeapAllocator::allocate(_size - sizeof(uintptr_t));
  }
}

void HeapAllocator::deallocate (void* _ptr)
{
  //
  // shift a pointer to a true value
  BlockInfo* p = true_ptr(_ptr);
  //
  // get physical length of allocated block
  size_t size = p->get_size();
  //
  // loop and glue adjacent free blocks
  do {
    //
    // get a pointer to appropriate adjacent block
//    BlockInfo* q = (sub_ptr(p, start) & size) != 0
//      ? ptr_sub(p, size)
//      : ptr_add(p, size);
    BlockInfo* q = (BlockInfo*)ptr_add(start, (sub_ptr(p, start) ^ size));
    //
    // if it is free...
    if (q->size == size) {
      //
      // delete it from list
      q->remove();
      //
      // obtain a pointer to combined free block
      p = (BlockInfo*)ptr_add(start, sub_ptr(p, start) & sub_ptr(q, start));
      //
      // advance a size
      size <<= 1;
    }
    //
    // if it is busy...
    else break;
  } while (size < current_size);
  //
  // insert it into appropriate free blocks list
  (lists + get_order(size))->insert(p, size);
}

void* HeapAllocator::_reallocate (
  void* _ptr, size_t _size, size_t _oldsize
)
{
  void* p = allocate(_size - sizeof(uintptr_t));
  memcpy(p, _ptr, _oldsize - sizeof(uintptr_t));
  deallocate(_ptr);
  return p;
}

void* HeapAllocator::reallocate (void* _ptr, size_t _size)
{
  size_t oldsize;
  //
  // shift a pointer to a true value
  BlockInfo* p = true_ptr(_ptr);
  //
  // advance allocated size by number of bytes reserved for length
  _size += sizeof(uintptr_t);
  _size = rounds[get_order(_size)];
  //
  // get physical length of allocated block
  oldsize = p->get_size();
  if (_size > oldsize) {
    if ((sub_ptr(p, start) & (_size - 1)) == 0) {
      size_t s = oldsize;
      while (s < _size) {
        if ((ptr_add(p, s))->size != s)
        {
          return _reallocate(_ptr, _size, oldsize);
        } else {
          (ptr_add(p, s))->remove();
          s <<= 1;
          p->set_size(s);
        }
      }
      return _ptr;
    } else {
      return _reallocate(_ptr, _size, oldsize);
    }
  } else if (_size < oldsize) {
    while (_size < oldsize) {
      oldsize >>= 1;
      (lists + get_order(oldsize))->insert(ptr_add(p, oldsize), oldsize);
    }
    p->set_size(_size);
    return _ptr;
  } else {
    return _ptr;
  }
}

#if DEBUG

#include <stdio.h>

void HeapAllocator::memory_dump ()
{
  {
    BlockInfo* p = (BlockInfo*)(start);
    printf("Heap at %p\n", p);
    do {
      if ((p->size & 0x01) != 0) {
        printf("a:%u ", p->get_size());
        p = ptr_add(p, p->get_size());
      } else {
        printf("f:%u ", p->get_size());
        p = ptr_add(p, p->get_size());
      }
    } while ((size_t)((char*)p - (char*)start) < current_size);
    printf("\n\n");
  }
  {
    unsigned i;
    BlockInfo* p;
    for (i = get_order(min_free_size);
         i <= get_order(current_size);
         i++)
    {
      BlockInfo* q;
      printf("%4u:", rounds[i]);
      p = lists + i;
      q = p->next;
      if (q != p) {
        do {
          printf(" %p(%p,%p)", q, q->prev, q->next);
          q = q->next;
        } while (q != p);
      }
      printf("\n");
    }
    printf("\n");
  }
}

#endif // DEBUG

}
