//
// Copyright (C) 2000, 2001 Andrey Slepuhin <pooh@msu.ru>
//
// libp++ is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// libp++ is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libp++; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source: /cvsroot/grace/src/pxx_ring.hh,v $
// $Revision: 1.10 $
// $Date: 2001/07/23 15:10:03 $
// Author: Andrey Slepuhin <pooh@msu.ru>

#ifndef __pxx_ring_hh__
#define __pxx_ring_hh__

#include "pxx_common.hh"
#include "pxx_spinlock.hh"
#include "pxx_sync.hh"
#include "pxx_i_chunk_allocator.hh"

namespace pxx
{

template <typename type_t> class Ring ;

template <typename type_t>
class RingNode
{

friend class Ring<type_t> ;

protected:

  RingNode* prev ;
  RingNode* next ;
  type_t* value ;

  inline RingNode (type_t* _value = null) :
    prev (this),
    next (this),
    value (_value)
  {}

  inline RingNode (RingNode* _prev, RingNode* _next, type_t* _value = null) :
    prev (_prev),
    next (_next),
    value (_value)
  {}

  inline ~RingNode ()
  {}

};

template <typename type_t>
class Ring :
  protected Spinlock,
  protected RingNode<type_t>
{

private:

  I_ChunkAllocator& allocator;
  size_t size;

public:

  Ring (I_ChunkAllocator& _allocator, type_t* _value = null) :
    RingNode<type_t> (_value),
    allocator (_allocator),
    size(0)
  {}

  ~Ring ()
  {}

  inline void put (type_t* _value)
  {
    RingNode<type_t>* p =
      (RingNode<type_t>*)allocator.allocate();
    #if PARANOIA
    if (p == 0)
      FATAL("Unable to allocate new ring node");
    #endif
    synchronize(Spinlock, self);
    new(p) RingNode<type_t>(prev, this, _value);
    prev = prev->next = p;
    size++;
  }

  inline void put_front (type_t* _value)
  {
    RingNode<type_t>* p =
      (RingNode<type_t>*)allocator.allocate();
    #if PARANOIA
    if (p == 0)
      FATAL("Unable to allocate new ring node");
    #endif
    synchronize(Spinlock, self);
    new(p) RingNode<type_t>(this, next, _value);
    next = next->prev = p;
    size++;
  }

  inline type_t* get ()
  {
    synchronize(Spinlock, self);
    if (prev != this) {
      return _remove(prev);
    } else {
      return 0;
    }
  }

  inline type_t* get_front ()
  {
    synchronize(Spinlock, self);
    if (next != this) {
      return _remove(next);
    } else {
      return 0;
    }
  }

  inline type_t* remove (RingNode<type_t>* _ptr)
  {
    synchronize(Spinlock, self);
    _remove(_ptr);
  }

  inline type_t* remove (type_t* _ptr)
  {
    synchronize(Spinlock, self);
    RingNode<type_t>* p = next;
    while (p != this) {
      if (p->value == _ptr) return _remove(p);
      p = p->next;
    }
    #if PARANOIA
    FATAL("Ring element not found");
    #endif
    return 0;
  }

  inline bool has (type_t* _ptr)
  {
    synchronize(Spinlock, self);
    RingNode<type_t>* p = next;
    while (p != this) {
      if (p->value == _ptr) return true;
      p = p->next;
    }
    return false;
  }

  inline bool is_empty ()
  {
    synchronize(Spinlock, self);
    return next == this;
  }

  inline size_t get_size ()
  {
    synchronize(Spinlock, self);
    return size;
  }

private:

  inline type_t* _remove (RingNode<type_t>* _ptr)
  {
    #if PARANOIA
    if (_ptr == this)
      FATAL("Attempt to remove fake ring node");
    #endif
    _ptr->prev->next = _ptr->next;
    _ptr->next->prev = _ptr->prev;
    type_t* p = _ptr->value;
    allocator.deallocate(_ptr);
    size--;
    return p;
  }

};

}

#endif // __pxx_ring_hh__
