#ifndef _TAUTOMAT_H_
#define _TAUTOMAT_H_ 


// ----------------------------------------------------------------------------

#include "t_stl.h"

// ----------------------------------------------------------------------------

using namespace std;

class GraphException {
	
  t_string desc_;

  public:
  
  GraphException(const char *desc) : desc_(desc) {}
  const t_string& get_desc() const { return desc_; }
};

template <class VKey, class VVal, class EdgeVal> class AttGraph {

  public:

  typedef t_map<VKey, VVal> Vertices;

  private:
	
  t_map<VKey, VVal> V;

  class EdgeInfo {

    public:	

    VKey to; EdgeVal label;

    EdgeInfo(const VKey& v, const EdgeVal l):to(v), label(l) {}
                  
    EdgeInfo() {}

    EdgeInfo& operator= (const EdgeInfo& ei) 
      { to=ei.to; label=ei.label; }

    // void *operator new(size_t sz) { return talloc(sz); }
  };

  t_multimap<VKey, EdgeInfo> E;
  t_multimap<VKey, EdgeInfo> E_rev_; // "" 

  public:

  // void *operator new(size_t sz) { return talloc(sz); }

  typedef VKey vertix_key;
  typedef t_multimap<VKey, EdgeInfo> Edges;
  typedef /* typename */ t_multimap<VKey, EdgeInfo>::const_iterator I;
  typedef /* typename */ t_multimap<VKey, EdgeInfo>::const_iterator const_EI;
  typedef /* typename */ t_map<VKey, VVal>::const_iterator const_VI;

  void
  addVertex(const VKey& vert, const VVal& val) 
    { V[vert]=val; }
	
  void
  addEdge(const VKey& vert1, const VKey& vert2, const EdgeVal eval)
  {
    E.insert(make_pair(vert1, EdgeInfo(vert2, eval)));
    E_rev_.insert(make_pair(vert2, EdgeInfo(vert1, eval)));
  }

  void
  removeEdge(const VKey& vert1, const VKey& vert2)
  {
    for(Edges::iterator tmp=E.lower_bound(vert1), i=E.end();
        tmp != E.upper_bound(vert1) && ((i=tmp++) != E.upper_bound(vert1));
        /* empty increment */)
    {
      if(i->second.to == vert2) E.erase(i);
    }

    for(Edges::iterator tmp=E_rev_.lower_bound(vert2), i=E_rev_.end();
        tmp != E_rev_.upper_bound(vert2) &&
        ((i=tmp++) != E_rev_.upper_bound(vert2));
        /* empty increment */)
    {
      if(i->second.to == vert1) E_rev_.erase(i);
    }
  }

  pair<I,I>
  getChildren(const VKey& vert) const
  {
    return E.equal_range(vert);
  }

  pair<I,I>
  getParents(const VKey& vert) const
  {
    return E_rev_.equal_range(vert);
  }
	
  bool
  isChild(const VKey& vert1, const VKey& vert2) const
  {
    pair<I,I> edges=getChildren(vert1);
    for(I i=edges.first; i!=edges.second; ++i)
      if(i->second.to == vert2) return true;
    return false;
  }
	
  EdgeVal
  getEdgeLabel(const VKey& vert1, const VKey& vert2) const
  {
    pair<I,I> edges=getChildren(vert1);
    for(I i=edges.first; i!=edges.second; ++i)
      if(i->second.to == vert2) return i->second.label;
    throw GraphException("Invalid edge"); 
  }

  const Edges&
  getEdges() const
  {
    return E;
  }

  const Vertices&
  getVertices() const {
    return V;
  }

  VVal&
  operator[] (const VKey& vert) {
    return V[vert];
  }

  const VVal&
  operator[] (const VKey& vert) const 
  {
    return V[vert];
  }

  bool
  isVertex(const VKey& vert) const 
  {
    if(V.find( vert) != V.end() )
      return true;
    return false;
  }

  AttGraph<VKey, VVal, EdgeVal>&
  operator= (const AttGraph<VKey, VVal, EdgeVal>& g2)
  {
    V = g2.V;
    E = g2.E;
    return *this;
  }

  // (18.07.2003)
  //
  // getFromLabels ---  ,    
  //                    f.
  //
  // getToLabels   ---  ,    
  //                    t.                   
 
  t_set<EdgeVal> getFromLabels(const VKey &f) const
  {
    t_set<EdgeVal> res;

    pair<I,I> edges = getChildren(f);
   
    for(I i = edges.first; i != edges.second; ++ i) 
      res.insert( i->second.label );	  
	    
    return res;
  }	  

  t_set<EdgeVal> getToLabels(const VKey &t) const
  {
    t_set<EdgeVal> res;
	  
    pair<I,I> edges = getParents(t);

    for(I i = edges.first; i != edges.second; ++ i)
      res.insert( i->second.label );	    
    
    return res;
  }	  
};

// ----------------------------------------------------------------------------

typedef long GraphVertexKey;
typedef long GraphVertexValue;
typedef long GraphLabelType;

typedef AttGraph<GraphVertexKey, GraphVertexValue, GraphLabelType> Graph;

// ----------------------------------------------------------------------------

template <class Q, class Sigma> class NFAInstance;

template <class Q, class Sigma>
class NFA
{
  AttGraph<Q, long, Sigma> G_; //  
  t_set<Q> currentStates_; //   
  t_set<Q> startStates_; //   
  t_set<Q> finalStates_; //   
  Sigma epsilon_;      //  epsilon-

  //
  //    
  //     currentStates_  
  // ,    epsilon-
  // 

  void buildClosure_()
  {
    t_queue<Q> todo;

    for(t_set<Q>::const_iterator i = currentStates_.begin(); 
        i != currentStates_.end(); 
        ++i) 
    {
      todo.push( *i );
    }
	
    while( !todo.empty() )
    {
      Q cstate = todo.front();

      pair<AttGraph<Q,long,Sigma>::I, 
           AttGraph<Q,long,Sigma>::I> 
      edges = G_.getChildren( cstate );
		
      for(AttGraph<Q,long,Sigma>::I e=edges.first;
          e != edges.second;
          ++e)
      {	
        if( e->second.label == epsilon_ &&
            currentStates_.find( e->second.to ) == 
            currentStates_.end() )
        {
          currentStates_.insert( e->second.to );
          todo.push( e->second.to );
        }
      }

      todo.pop();
    }	
  }

  public:
	
  NFA(const Sigma& epsilon): 
    epsilon_(epsilon) 
  { }

  // void *operator new(size_t sz) { return talloc(sz); }

  void setCurrentStates(const t_set<Q> qs)
  {
    currentStates_ = qs; buildClosure_();
  }

  void restoreCurrentStates(const t_set<Q> qs)
  {
    currentStates_ = qs;
  }

  void addCurrentState(const Q& q)
  {
    currentStates_.insert( q );
    buildClosure_();
  }

  //
  //   
  //

  void addState(const Q& q) 
    { G_.addVertex(q, 0); }

  void addStartState(const Q& q)
    { startStates_.insert(q); }
	
  void addFinalState(const Q& q) 
    { finalStates_.insert(q); }

  void addTransition(const Q& q1, const Q& q2, const Sigma& s) 
  {
    G_.addEdge(q1,q2,s);
  }

  //
  // "" .     
  //

  void readSymbol(const Sigma& sym)
  {
    t_set<Q> newCurrentStates;
		
    for(t_set<Q>::const_iterator i = currentStates_.begin();
	i != currentStates_.end();
	++i)
    {
      pair<AttGraph<Q,long,Sigma>::I, 
           AttGraph<Q,long,Sigma>::I> 
      edges = G_.getChildren( *i );
                
      for(AttGraph<Q,long,Sigma>::I e=edges.first;
          e != edges.second;
          ++e)
      {
	if( e->second.label == sym )
	  newCurrentStates.insert( e->second.to );
      }
    }

    setCurrentStates( newCurrentStates );	
  }

  const t_set<Q>& getCurrentStates() const { return currentStates_; }

  const t_set<Q>& getStartStates() const { return startStates_; }

  const t_set<Q>& getFinalStates() const { return finalStates_; }

  const AttGraph<Q,long,Sigma>::Edges& getTransitions() const {
    return G_.getEdges();
  }

  Sigma epsilon() const { return epsilon_; }
	
  NFA<Q, Sigma>&
  operator=(const NFA<Q, Sigma>& nfa2)
  {
    G_ = nfa2.G_;
    currentStates_ = nfa2.currentStates_; //   
    startStates_ = nfa2.startStates_; //   
    finalStates_ = nfa2.finalStates_; //   
    epsilon_ = nfa2.epsilon_;      //  epsilon-

    return *this;
  }

  // ,     - 

  bool isInFinalState() const
  {
    for(t_set<Q>::const_iterator i = currentStates_.begin();
        i != currentStates_.end();
        ++i)
    {
      if( finalStates_.find( *i ) != finalStates_.end() )
	return true;
    }
		        
    return false;
  }	

  //
  //        
  //

  const t_set<Sigma> getValidTransitions() const
  {
    t_set<Sigma> res;

    // for each state

    for(t_set<Q>::const_iterator i = currentStates_.begin();
        i != currentStates_.end();
        ++i)
    {
      // for each edge from i'th state

      pair<AttGraph<Q,long,Sigma>::I, 
           AttGraph<Q,long,Sigma>::I> 
      edges = G_.getChildren( *i );
		
      for(AttGraph<Q,long,Sigma>::I e=edges.first;
          e != edges.second;
          ++e)
      {
        if( e->second.label != epsilon_ )
          res.insert( e->second.label );
      }
    }

    return res;
  }

  void reset() { currentStates_ = startStates_; }

  friend class NFAInstance<Q,Sigma>;

  // (18.07.2003)
  //
  //        
  // (  getValidTransitions   NFAInstance)
  
  //
  // (28.07.2003)
  //
  
  public:

  const t_set<Sigma> getStartLabels() const
  {	  
    t_set<Sigma> res;
	  
    //    

    t_set<Q>::const_iterator i;   
    for(i = startStates_.begin(); i != startStates_.end(); i ++) 
    {
      //     i- 
  
      pair<AttGraph<Q,long,Sigma>::I, AttGraph<Q,long,Sigma>::I> edges = 
        G_.getChildren( *i );

      AttGraph<Q,long,Sigma>::I e;
      for(e = edges.first; e != edges.second; e ++)
        if( e->second.label != epsilon_ ) 
          res.insert( e->second.label );
    }

    return res;
  }	  
};	

template <class Q, class Sigma>
class NFAInstance // : private NFA<Q,Sigma>
{
  const NFA<Q,Sigma>& nfa_;
  t_set<Q> currentStates_; //   

  void buildClosure_()
  {
    t_queue<Q> todo;
	
    for(t_set<Q>::const_iterator i = currentStates_.begin(); 
        i != currentStates_.end(); 
        ++i) 
    {
      todo.push( *i );
    }
	
    while( !todo.empty() )
    {
      Q cstate = todo.front();

      pair<AttGraph<Q,long,Sigma>::I, 
           AttGraph<Q,long,Sigma>::I> 
      edges = nfa_.G_.getChildren( cstate );
		
      for(AttGraph<Q,long,Sigma>::I e=edges.first;
          e != edges.second;
          ++e)
      {	
	if( e->second.label == nfa_.epsilon_ &&
	    currentStates_.find( e->second.to ) == 
	    currentStates_.end() )
	{
	  currentStates_.insert( e->second.to );
	  todo.push( e->second.to );
	}
      }
			
      todo.pop();
    }	
  }
  
  public:

  NFAInstance(const NFA<Q,Sigma> &nfa):
    nfa_(nfa)
  {
    reset();
  }
  
  void reset() { currentStates_ = nfa_.startStates_; }

  const t_set<Q>& getCurrentStates() const { return currentStates_; }

  //
  // ,     -
  // 
  //

  bool isInFinalState() const
  {
    for(t_set<Q>::const_iterator i = currentStates_.begin();
        i != currentStates_.end();
        ++i)
    {
      if( nfa_.finalStates_.find( *i ) != nfa_.finalStates_.end() )
	return true;
    }
		        
    return false;
  }	

  void setCurrentStates(const t_set<Q> qs)
  {
    currentStates_ = qs; buildClosure_();
  }

  void restoreCurrentStates(const t_set<Q> qs)
  {
    currentStates_ = qs;
  }

  void addCurrentState(const Q& q)
  {
    currentStates_.insert( q );
    buildClosure_();
  }

  //
  // "" .     
  //

  void readSymbol(const Sigma& sym)
  {
    t_set<Q> newCurrentStates;
		
    for(t_set<Q>::const_iterator i = currentStates_.begin();
        i != currentStates_.end();
        ++i)
    {
      // for each edge from i'th state

      pair<AttGraph<Q,long,Sigma>::I, 
           AttGraph<Q,long,Sigma>::I> 
      edges = nfa_.G_.getChildren( *i );
                
      for(AttGraph<Q,long,Sigma>::I e=edges.first;
          e != edges.second;
          ++e)
      {
	if( e->second.label == sym )
	  newCurrentStates.insert( e->second.to );
      }
    }

    setCurrentStates( newCurrentStates );	
  }

  //
  //        
  //
  
  const t_set<Sigma> getValidTransitions() const
  {
    t_set<Sigma> res;

    // for each state

    for(t_set<Q>::const_iterator i = currentStates_.begin();
        i != currentStates_.end();
        ++ i)
    {
      // for each edge from i'th state

      pair<AttGraph<Q,long,Sigma>::I, 
           AttGraph<Q,long,Sigma>::I> 
      edges = nfa_.G_.getChildren( *i );
		
      for(AttGraph<Q,long,Sigma>::I e=edges.first;
          e != edges.second;
          ++ e)
      {
	if( e->second.label != nfa_.epsilon_ )
	  res.insert( e->second.label );
      }
    }

    return res;
  }

  void print() const {}
};

typedef long NFAState;
typedef t_set<NFAState> NFAStatesSet;
typedef NFA<NFAState, GraphLabelType> IntNFA;
typedef NFAInstance<NFAState, GraphLabelType> IntNFAInstance;

// (07.07.2003)

typedef AttGraph<NFAState, long, GraphLabelType>::Edges IntNFATransitions;

// ----------------------------------------------------------------------------

struct QueryEdge 
{
  long    from;
  long    to;
  IntNFA* label;
  IntNFA* reverse_label; // (11.07.2003) ---    

  QueryEdge(): label(NULL),
               reverse_label(NULL)          
    {}	  
};

typedef t_list<QueryEdge> Query;

// ----------------------------------------------------------------------------

// (25.07.2003)
// constraint (local vicinity constraint)

struct ConstraintEdge
{
  long          from;
  NFAStatesSet* label;
  NFAStatesSet* reverse_label;
  
  ConstraintEdge(): label(NULL),
                    reverse_label(NULL)
    {}	  
};	

typedef t_list<ConstraintEdge> Constraint;

// ----------------------------------------------------------------------------

t_set<GraphVertexKey> wave_regexp_closure(
  const Graph &G, IntNFAInstance &M, const GraphVertexKey &s);

// (14.07.2003)
//
//   wave_regexp_closure 

t_set<GraphVertexKey> wave_regexp_closure_r(
  const Graph &G, IntNFAInstance &M, const GraphVertexKey &s);

// ----------------------------------------------------------------------------

#endif // end of _TAUTOMAT_H_
