#ifndef _AUTOMAT_H_
#define _AUTOMAT_H_

// ----------------------------------------------------------------------------

#include <string>
#include <set>
#include <map>
#include <list>
#include <iostream>
#include <fstream>

using namespace std;

// ----------------------------------------------------------------------------

class GraphException {
	
  string desc_;

  public:
  
  GraphException(const char *desc) : desc_(desc) {}
  const string& get_desc() const { return desc_; }
};

template <class VKey, class VVal, class EdgeVal> class AttGraph {

  public:

  typedef map<VKey, VVal> Vertices;

  private:
	
  map<VKey, VVal> V;

  class EdgeInfo {

    public:	

    VKey to; EdgeVal label;

    EdgeInfo(const VKey& v, const EdgeVal l):to(v), label(l) {}
                  
    EdgeInfo() {}

    EdgeInfo& operator= (const EdgeInfo& ei) 
      { to=ei.to; label=ei.label; }

    void *operator new(size_t sz) { return talloc(sz); }
  };

  multimap<VKey, EdgeInfo> E;
  multimap<VKey, EdgeInfo> E_rev_; // "" 

  public:

  void *operator new(size_t sz) { return talloc(sz); }

  typedef VKey vertix_key;
  typedef multimap<VKey, EdgeInfo> Edges;
  typedef typename multimap<VKey, EdgeInfo>::const_iterator I;
  typedef typename multimap<VKey, EdgeInfo>::const_iterator const_EI;
  typedef typename map<VKey, VVal>::const_iterator const_VI;

  void
  addVertex(const VKey& vert, const VVal& val) 
    { V[vert]=val; }
	
  void
  addEdge(const VKey& vert1, const VKey& vert2, const EdgeVal eval)
  {
    E.insert(make_pair(vert1, EdgeInfo(vert2, eval)));
    E_rev_.insert(make_pair(vert2, EdgeInfo(vert1, eval)));
  }

  void
  removeEdge(const VKey& vert1, const VKey& vert2)
  {
    for(Edges::iterator tmp=E.lower_bound(vert1), i=E.end();
        tmp != E.upper_bound(vert1) && ((i=tmp++) != E.upper_bound(vert1));
        /* empty increment */)
    {
      if(i->second.to == vert2) E.erase(i);
    }

    for(Edges::iterator tmp=E_rev_.lower_bound(vert2), i=E_rev_.end();
        tmp != E_rev_.upper_bound(vert2) &&
        ((i=tmp++) != E_rev_.upper_bound(vert2));
        /* empty increment */)
    {
      if(i->second.to == vert1) E_rev_.erase(i);
    }
  }

  pair<I,I>
  getChildren(const VKey& vert) const
  {
    return E.equal_range(vert);
  }

  pair<I,I>
  getParents(const VKey& vert) const
  {
    return E_rev_.equal_range(vert);
  }
	
  bool
  isChild(const VKey& vert1, const VKey& vert2) const
  {
    pair<I,I> edges=getChildren(vert1);
    for(I i=edges.first; i!=edges.second; ++i)
      if(i->second.to == vert2) return true;
    return false;
  }
	
  EdgeVal
  getEdgeLabel(const VKey& vert1, const VKey& vert2) const
  {
    pair<I,I> edges=getChildren(vert1);
    for(I i=edges.first; i!=edges.second; ++i)
      if(i->second.to == vert2) return i->second.label;
    throw GraphException("Invalid edge"); 
  }

  const Edges&
  getEdges() const
  {
    return E;
  }

  const Vertices&
  getVertices() const {
    return V;
  }

  VVal&
  operator[] (const VKey& vert) {
    return V[vert];
  }

  const VVal&
  operator[] (const VKey& vert) const 
  {
    return V[vert];
  }

  bool
  isVertex(const VKey& vert) const 
  {
    if(V.find( vert) != V.end() )
      return true;
    return false;
  }

  AttGraph<VKey, VVal, EdgeVal>&
  operator= (const AttGraph<VKey, VVal, EdgeVal>& g2)
  {
    V = g2.V;
    E = g2.E;
    return *this;
  }
};

typedef long GraphVertexKey;
typedef long GraphVertexValue;
typedef long GraphLabelType;

typedef AttGraph<GraphVertexKey, GraphVertexValue, GraphLabelType> Graph;

// ----------------------------------------------------------------------------

template <class Q, class Sigma> class NFAInstance;

template <class Q, class Sigma>
class NFA
{
  AttGraph<Q, long, Sigma> G_; //  
  set<Q> currentStates_; //   
  set<Q> startStates_; //   
  set<Q> finalStates_; //   
  Sigma epsilon_;      //  epsilon-
	
  public:
	
  NFA(const Sigma& epsilon) : epsilon_(epsilon) {}

  //
  //   
  //

  void addState(const Q& q) 
    { G_.addVertex(q, 0); }

  void addStartState(const Q& q)
    { startStates_.insert(q); }
	
  void addFinalState(const Q& q) 
    { finalStates_.insert(q); }

  void addTransition(const Q& q1, const Q& q2, const Sigma& s) 
  {
    G_.addEdge(q1,q2,s);
  }

  const set<Q>& getStartStates() const 
    { return startStates_; }
  
  const set<Q>& getFinalStates() const 
    { return finalStates_; }

  const AttGraph<Q,long,Sigma>::Edges& getTransitions() const 
  {
    return G_.getEdges();
  }

  Sigma epsilon() 
    { return epsilon_; }
	
  void iterate()
  {
    for(set<Q>::const_iterator i = finalStates_.begin();
        i != finalStates_.end();
        ++i)
    {
      for(set<Q>::const_iterator j = startStates_.begin();
          j != startStates_.end();
          ++j)
      {
        addTransition(*i, *j, epsilon_);
        addTransition(*j, *i, epsilon_);
      }
    }
  }
  
  void alternate(const NFA<Q, Sigma>& m2)
  {
    Q max_state = this->getMaxState()+1;
    Q new_start;
    Q new_end;
    const map<Q, long>& m2states = m2.G_.getVertices();
		
    new_start = max_state;
    addState( max_state++ );

    new_end = max_state;
    addState( max_state++ );

    for(map<Q, long>::const_iterator i = m2states.begin();
        i != m2states.end();
        ++i)
    {
      addState( max_state + i->first );
    }

    for(map<Q, long>::const_iterator i = m2states.begin();
        i != m2states.end();
        ++i)
    {
      pair<AttGraph<Q,long,Sigma>::I, 
           AttGraph<Q,long,Sigma>::I> 
      edges = m2.G_.getChildren( i->first );
	
      for(AttGraph<Q,long,Sigma>::I e=edges.first;
          e != edges.second;
          ++e)
      {
        addTransition(max_state + e->first,
                      max_state + e->second.to,
                      e->second.label);
      }
    }

    for(set<Q>::const_iterator i = startStates_.begin();
        i != startStates_.end();
        ++i)
    {
      addTransition(new_start, *i, epsilon_);
    }
    
    for(set<Q>::const_iterator i = m2.startStates_.begin();
        i != m2.startStates_.end();
        ++i)
    {
      addTransition(new_start, *i+max_state, epsilon_);
    }

    startStates_.clear();
    startStates_.insert(new_start);

    for(set<Q>::const_iterator i = finalStates_.begin();
        i != finalStates_.end();
        ++i)
    {
      addTransition(*i, new_end, epsilon_);
    }
    
    for(set<Q>::const_iterator i = m2.finalStates_.begin();
        i != m2.finalStates_.end();
        ++i)
    {
      addTransition(*i+max_state, new_end, epsilon_);
    }
    
    finalStates_.clear();
    finalStates_.insert(new_end);
  }
 
  Q getMaxState() 
  { 
    bool first_state = true;
    Q max_state;
    const map<Q, long>& allstates = G_.getVertices();

    for(map<Q, long>::const_iterator i = allstates.begin();
	i != allstates.end();
	++i)
    {
       if( first_state ) {
	 max_state = i->first;
	 first_state = false;
       }
       else if( i->first > max_state )
	 max_state = i->first;
    }
		
    if( first_state ) throw "No states found at all!";
    
    return max_state;
  }
  
  void concat(const NFA<Q, Sigma>& m2)
  {
    Q max_state = this->getMaxState()+1;
    const map<Q, long>& m2states = m2.G_.getVertices();
		
    for(map<Q, long>::const_iterator i = m2states.begin();
        i != m2states.end();
        ++i)
    {
      addState( max_state + i->first );
    }

    for(map<Q, long>::const_iterator i = m2states.begin();
        i != m2states.end();
        ++i)
    {
      pair<AttGraph<Q,long,Sigma>::I, 
           AttGraph<Q,long,Sigma>::I> 
      edges = m2.G_.getChildren( i->first );
	
      for(AttGraph<Q,long,Sigma>::I e=edges.first;
          e != edges.second;
          ++e)
      {
        addTransition(max_state + e->first,
                      max_state + e->second.to,
                      e->second.label);
      }
    }

    for(set<Q>::const_iterator i = finalStates_.begin();
        i != finalStates_.end();
        ++i)
    {
      for(set<Q>::const_iterator i2 = m2.startStates_.begin();
          i2 != m2.startStates_.end();
          ++i2)
      {
        addTransition(*i, *i2+max_state, epsilon_);
      }
    }
    
    finalStates_.clear();
    
    for(set<Q>::const_iterator i = m2.finalStates_.begin();
        i != m2.finalStates_.end();
        ++i)
    {
      finalStates_.insert(max_state + *i);
    }
  }
};	

typedef long NFAState;
typedef NFA<NFAState, GraphLabelType> IntNFA;
typedef NFA<NFAState, string>         StrNFA;

// ----------------------------------------------------------------------------

class GenerateAutomat {

  bool error_; StrNFA M_;

  public:

  GenerateAutomat(const char* reg_expr);

 ~GenerateAutomat() {
  }	 

  void set_error() {
    error_ = true;
  }	  

  bool get_error() {
    return error_;
  }	  

  void set_result(const StrNFA &M) {
     M_ = M;
  }	  

  const StrNFA &result() const {
    return  M_;
  }	  
};

enum TokenType {
  STAR, PLUS, QUESTION, VERTICALBAR, LEFT, RIGHT, CHAR, END
};

struct Token {
  TokenType type; string value;

  Token();
  Token(TokenType t, const char *v);
  Token &operator= (const Token &t);
  bool operator== (const Token &t) const;
  bool operator!= (const Token &t) const;
};

class Lex {

  string buffer; unsigned int cursor;

  protected:

  const Token star;
  const Token plus;
  const Token question;
  const Token verticalbar;
  const Token left;
  const Token right;
  const Token end;

  public:

  Lex(const char *str);

  const Token current_lexem();
  const Token push_lexem();
};

class Parser: protected Lex
{
  typedef StrNFA* RegExpNFA_ptr;

  void treat_atom(RegExpNFA_ptr result);
  void treat_Atom(RegExpNFA_ptr result);
  void treat_sequence(RegExpNFA_ptr result);
  void treat_regular(RegExpNFA_ptr result);

  GenerateAutomat *generate_automat;

  public:

  typedef int Exception;

  Parser(const char *str, GenerateAutomat &ga);

  void start() throw(Exception);
};

// ----------------------------------------------------------------------------

struct LocCode {
  long   code;
  string word;
};

typedef list<LocCode> LocDictionary;

long add_loc_code(LocDictionary &, string);

// ----------------------------------------------------------------------------

struct QueryEdge {
  long   from;
  long   to;
  string label;
};

typedef list<QueryEdge> Query;

// ----------------------------------------------------------------------------

#endif // end of _AUTOMAT_H_
