#include "geometry.h"
#include "tracer.h"

struct texture def_tex={0.7,0.5,0.5,0, {1,1,1}, {1.8,0}, 10};

void sphere_set (struct solid *s, const struct vector *v, const double r)
{
	s->type = 0;		// sphere
	s->obj[0].sphere.center = *v;
	s->obj[0].sphere.radius = r;
}

void plane_set (struct solid *s, const struct vector *n, const double d)
{
	s->type = 1;		// plane
	s->obj[0].plane.dist = d;
	s->obj[0].plane.norm = *n;
}

void poly_set (struct solid *s, const int n)
{
	struct vector	v1, v2;
	struct vector	cross;
	struct poly	*p;
	int		count;
	
	p=&(s->obj[0].poly);
	s->type = 2;		// poly
	p->np = n;

	// FIXME v1 & v2 shouldn't be collinear
	sub_v_v (&v1, p->vertex[0].point, p->vertex[1].point);
	sub_v_v (&v2, p->vertex[2].point, p->vertex[1].point);
	vmult_v_v (&cross, v1, v2);
	normalyze (&(p->plane.norm), cross);
	p->plane.dist = -dot_v_v (p->plane.norm, p->vertex[0].point);
	for (count = 0; count < n-1; count++) {
		sub_v_v (&v1,p->vertex[count+1].point,p->vertex[count].point);
		vmult_v_v (&(p->vertex[count].norm), v1, p->plane.norm);
	}
	sub_v_v (&v1,p->vertex[0].point,p->vertex[n-1].point);
	vmult_v_v (&(p->vertex[n-1].norm), v1, p->plane.norm);
}

int sphere_isect (const union gobj *obj, const struct ray *r, double *dist)
{
	struct vector cen, l;
	double rad, l2, tc, t2;
	double res;

	cen = obj->sphere.center;
	rad = obj->sphere.radius;

	sub_v_v (&l, cen, r->org);
	l2 = dot_v_v (l, l);
	tc = dot_v_v (l, r->dir);
	t2 = rad * rad - l2 + tc * tc;
	if (t2 <= 0)
		return 0;
	t2 = sqrt (t2);

	if (tc < t2) {
		res = tc + t2;
		//t2    = tc - t2;
	} else {
		res = tc - t2;
		//t2    = tc + t2;
	}

	*dist = res;
	return (res > 1e-5);
}

int sphere_norm (const union gobj *obj, const struct vector *p,
		 struct vector *v)
{
	struct vector c;
	double r;

	c = obj->sphere.center;
	r = obj->sphere.radius;

	v->x = (p->x - c.x) / r;
	v->y = (p->y - c.y) / r;
	v->z = (p->z - c.z) / r;

	return 1;
}

int plane_isect (const union gobj *obj, const struct ray *r, double *dist)
{
	double d, vd;
	struct vector n;
	double res;

	d = obj->plane.dist;
	n = obj->plane.norm;

	vd = dot_v_v (n, r->dir);
	res = -(dot_v_v (n, r->org) + d) / vd;

	*dist = res;

	return (res > 1e-5);
}

int plane_norm (const union gobj *obj, const struct vector *p, struct vector *v)
{
	*v = obj->plane.norm;
	return 1;
}

int poly_isect (const union gobj *obj, const struct ray *r, double *dist)
{
	double d, vd;
	struct vector n;
	int np;
	double res;

	d = obj->poly.plane.dist;
	n = obj->poly.plane.norm;
	np = obj->poly.np;

	vd = dot_v_v (n, r->dir);
	*dist = res = -(dot_v_v (n, r->org) + d) / vd;

	if (res > 1e-5) {
		int count;
		struct vector tmpv, isect;

		isect_point (&isect, *r, res);
		for (count = 0; count < np; count++) {
			sub_v_v (&tmpv, isect, obj->poly.vertex[count].point);
			if (dot_v_v (obj->poly.vertex[count].norm, tmpv) < 1e-5) {
				return 0;
			}
		}
		return 1;
	}

	return 0;
}

int poly_norm (const union gobj *obj, const struct vector *p, struct vector *v)
{
	*v = obj->plane.norm;
	return 1;
}

struct solid_vmt vmt[NUM_OBJECTS] = {
	{sphere_isect, sphere_norm},
	{plane_isect, plane_norm},
	{poly_isect, poly_norm},
};

double plight_shadow (const struct scene *s, const struct plight *pl,
		const struct vector *p, struct vector *l)
{
	double dist;
	double attenuation;
	double t;
	struct ray r;
	struct texture *texture;
	const struct solid *occlude;

	l->x = (pl->location).x - p->x;
	l->y = (pl->location).y - p->y;
	l->z = (pl->location).z - p->z;
	dist = sqrt ((l->x) * (l->x) + (l->y) * (l->y) + (l->z) * (l->z));
	attenuation = (pl->distscale) / dist;
	l->x /= dist;
	l->y /= dist;
	l->z /= dist;
	setr_v_v (&r, *p, *l);

	attenuation = attenuation * attenuation;
	// FIXME: _very_ slow
	while ((occlude = s_scene_isect (s, &r, &t)) != 0 && dist > t) {
		r.org.x = r.org.x + r.dir.x * t;
		r.org.y = r.org.y + r.dir.y * t;
		r.org.z = r.org.z + r.dir.z * t;
		texture = GET_TEXTURE(s, occlude->texture);
		//(*(occlude->get_tex)) (&s->scs, &(occlude->g), r.org, &texture);
		if (texture->kt < s->scs.threshold)
			return 0;
		if ((attenuation *= texture->kt) < s->scs.threshold)
			return 0;
		dist -= t;
	}
	return attenuation;
}

void set_plight (struct plight *pl, const struct vector *l,
		const struct vector *c, const double d)
{
	pl->color     = *c;
	pl->location  = *l;
	pl->distscale = d;
}
