#ifndef _RT_GEOMETRY_H
#define _RT_GEOMETRY_H

#include "vector.h"

#ifdef __cplusplus
extern "C" {
#endif


#define NUM_OBJECTS	3	// # object types

#define SCENE_SIZE	(1024*1024) // FIXME!!!

#define MAX_LIGHTS	8

struct plight {
	struct vector	color;
	struct vector	location;
	double		distscale;	// Wtf?
};

struct medium {
	double n_refr;		/* refraction coef. */
	double beta;
};

struct texture {
	double ka;		// ambient coef.
	double kd;		// diffuse coef.
	double ks;		// specular coef.
	double kt;		// transmission coef.
	struct vector color;	// color
	struct medium med;	// medium properties
	int p;			// phong's coef.
	//struct vector n;	// surface normal direction
};

struct sc_set {
	struct		vector eye, eyedir, vx, vy;
	struct		vector background;
	struct		medium air, glass;
	double		threshold;
	double		geomthreshold;
	struct		vector ambient;
	double		f_x_start;
	double		f_y_start;
	int		maxlevel;
	int		nx, ny;
};

struct scene {
	int		lights_count;
	int		solids_count;
	int		textures_count;
	int		solid_index;
	int		first_solid;
	int		size;
	struct sc_set	scs;
	struct plight	light[MAX_LIGHTS];
	char		pool[sizeof (int) + SCENE_SIZE];
};

struct sphere {
	double radius;
	struct vector center;
};

struct plane {
	double dist;
	struct vector norm;
};

struct vpoly {
	struct vector point;
	struct vector norm;
};

struct poly {
	struct plane plane;
	int np;
	struct vpoly vertex[0];
};

union gobj {
	struct sphere sphere;
	struct plane plane;
	struct poly poly;
};

struct solid {
	int next;		// next solid index
	int type;		// solid type (vmt index)
	int texture;		// texture index
	union gobj obj[0];	// the object's place
};

struct solid_vmt {
	int (*isect) (const union gobj *obj,	/*const struct sc_settings*, */
		      const struct ray *r, double *dist);
	int (*norm) (const union gobj *obj,	/*const struct sc_settings*, */
		     const struct vector*, struct vector *);
};

extern struct solid_vmt vmt[NUM_OBJECTS];

#define FIRST_SOLID(s)    (((s)->first_solid)? \
		(struct solid *)(&((s)->pool[(s)->first_solid])):0)

static inline void scene_init (struct scene *s)
{
	s->lights_count   = 0;
	s->solids_count   = 0;
	s->textures_count = 0;
	s->solid_index    = -1;
	s->first_solid    = -1;
	s->size           = 0;
}

static inline struct solid *next_solid (const struct scene *s,
		const struct solid *p)
{
	int next;

	next = p->next;
	return (next) ? (struct solid *) (s->pool + next) : 0;
}

static inline struct solid *alloc_solid (struct scene *s, const int size)
{
	struct solid *p;
	int sz;

	sz = s->size;
	if (sz + size > SCENE_SIZE)
		return 0;

	if (s->solid_index >= 0) {
		((struct solid *) (s->pool + s->solid_index))->next = sz;
	} else {
		s->first_solid = sz;
	}
	s->solid_index = sz;
	
	p = (struct solid *) (&(s->pool[sz]));
	p->next = 0;
	s->size += size;
	
	s->solids_count++;
	
	return p;
}

static inline int alloc_texture (struct scene *s, const int size)
{
	int sz;

	sz = s->size;
	if (sz + size > SCENE_SIZE)
		return -1;
	
	s->textures_count++;
	s->size += size;

	return sz;
}

#define SPHERE_SIZE	(sizeof(struct solid)+sizeof(struct sphere))
#define PLANE_SIZE	(sizeof(struct solid)+sizeof(struct plane))
#define POLY_SIZE(n)	(sizeof(struct solid)+sizeof(struct poly) \
			+n*sizeof(struct vpoly))
#define TEXTURE_SIZE	(sizeof(struct texture))
	
void sphere_set (struct solid *s, const struct vector*, const double);
void plane_set (struct solid *s, const struct vector*, const double);
void poly_set (struct solid *s, const int);

double plight_shadow (const struct scene*, const struct plight*,
		const struct vector*, struct vector*);

extern struct texture def_tex;
//	double ka;		// ambient coef.
//	double kd;		// diffuse coef.
//	double ks;		// specular coef.
//	double kt;		// transmission coef.
//	struct vector color;	// color
//	struct medium med;	// medium properties
//	int p;			// phong's coef.
//	//struct vector n;	// surface normal direction

#define GET_TEXTURE(s,t)	(((t)<0)?(&def_tex):(struct texture *)((s)->pool + (t)))

void set_plight (struct plight *pl, const struct vector *l,
		const struct vector *c, const double d);

#ifdef __cplusplus
};
#endif

#endif

