#include <stdio.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include "output.h"

struct out_handler {
	int (*stream_open)(struct out_descr *desc);
	int (*stream_close)(struct out_descr *desc);
	int (*frame_open)(struct out_descr *desc);
	int (*frame_close)(struct out_descr *desc);
	int (*out_fragment)(struct out_descr *desc, struct rect r, const struct pixel *px);
};

// TGA file output

struct tga_header {
	unsigned char magic[12]; /* {0,0,2,0,0,0,0,0,0,0,0,0}; */
	short int nx, ny;
	short int bits;
};

static int tga_stream_open (struct out_descr *desc)
{
	fprintf (stderr, "Open TGA output\n");
	return 0;
}

static int tga_stream_close (struct out_descr *desc)
{
	fprintf (stderr, "Close TGA output\n");
	return 0;
}

static int tga_frame_open (struct out_descr *desc)
{
	struct tga_header th;
	char name [32];
	int fd;

	if (desc->frames > 1)
		snprintf (name, 32, "%s_%03d.tga", desc->name, desc->frame);
	else
		snprintf (name, 32, "%s.tga", desc->name);
	fprintf(stderr, "Begin '%s'\n", name);
	
	fd = creat (name, 0600);
	if (fd == -1) return -1;

	memset (&th, 0, sizeof (th));
	th.magic[2] = 2;
	th.bits = 24;
	th.nx = desc->width;
	th.ny = desc->height;
	write (fd, &th, sizeof(th));
	
	desc->fd = fd;

	return fd;
}

static int tga_frame_close (struct out_descr *desc)
{
	int	res = 0;
	
	if (desc->fd != -1) {
		res = close (desc->fd);
		desc->fd = -1;
	}
	
	return res;
}

static int tga_out_fragment (struct out_descr *desc, struct rect r, const struct pixel *px)
{
	int y;

	for (y = 0; y < r.height; y++) {
		int pos;
		int fgmtsize;
		
		pos = sizeof(struct tga_header) + ((y+r.top)*(desc->width)+r.left)*sizeof(struct pixel);
		fgmtsize = r.width*sizeof(struct pixel);
		
		lseek (desc->fd, pos, SEEK_SET);
		if (write (desc->fd, px, fgmtsize) != fgmtsize) {
			//error (E_OFILE);
		}
		px += r.width;
	}

	return 0;
}

static struct out_handler tga_handler = {
	stream_open:	tga_stream_open,
	stream_close:	tga_stream_close,
	frame_open:	tga_frame_open,
	frame_close:	tga_frame_close,
	out_fragment:	tga_out_fragment,
};

// sock output

struct sock_cmd {
	unsigned short	sign, cmd;
	union {
		struct{
			int	frames;
		} start;
		struct{
			int	w,h;
		} dim;
		struct{
			int	l,t;
			int	w,h;
		} frag;
	} data;
};

#define SIGN	0x7991
enum {_START, _FRAG, _END, _FR_OPEN, _FR_CLOSE};


static int sock_stream_open (struct out_descr *desc)
{
	struct sockaddr_in saddr;
	struct hostent *hep;
	char hostname[1024];
	int hostport;
	int fd;
	int len;
	char *tmp;
	struct sock_cmd cmd;

	fprintf (stderr, "Open SOCKET output '%s'\n", desc->name);

	tmp = strchr (desc->name, ':');
	if (!tmp) return -1;
	fprintf(stderr, "name = %p\ntmp=%p\n", desc->name, tmp);
	hostport = atoi (tmp+1);
	memcpy (hostname, desc->name, tmp-desc->name);
	hostname[tmp-desc->name] = 0;

	fprintf(stderr, "connect to '%s:%d'\n", hostname, hostport);

	fd = socket (PF_INET, SOCK_STREAM, 0);
	if (fd < 0) {
		return -1;
	}
	
	len = 1;
	setsockopt (fd, SOL_SOCKET, SO_REUSEADDR, &len, sizeof(int));
	len = sizeof (saddr);
	memset (&saddr, 0, len);
	saddr.sin_family = AF_INET;
	saddr.sin_port = htons (hostport);
	hep = gethostbyname (hostname);
	memcpy (&saddr.sin_addr.s_addr, hep->h_addr, hep->h_length);
	
	if (connect (fd, (struct sockaddr*)&saddr, len) == -1) {
		close (fd);
		return -1;
	}

	memset (&cmd, 0, sizeof (cmd));
	cmd.sign = SIGN;
	cmd.cmd = _START;
	cmd.data.start.frames = desc->frames;
	cmd.data.dim.w = desc->width;
	cmd.data.dim.h = desc->height;
	write (fd, &cmd, sizeof (cmd));
	
	desc->fd = fd;
	
	return fd;
}

static int sock_stream_close (struct out_descr *desc)
{
	struct sock_cmd cmd;
	int res;
	
	memset (&cmd, 0, sizeof (cmd));
	cmd.sign = SIGN;
	cmd.cmd = _END;
	write (desc->fd, &cmd, sizeof (cmd));

	if (desc->fd != -1) res = close (desc->fd);
	desc->fd = -1;

	return res;
}

static int sock_frame_open (struct out_descr *desc)
{
	struct sock_cmd cmd;
	
	memset (&cmd, 0, sizeof (cmd));
	cmd.sign = SIGN;
	cmd.cmd = _FR_OPEN;
	return write (desc->fd, &cmd, sizeof (cmd));
}

static int sock_frame_close (struct out_descr *desc)
{
	struct sock_cmd cmd;
	
	memset (&cmd, 0, sizeof (cmd));
	cmd.sign = SIGN;
	cmd.cmd = _FR_CLOSE;
	return write (desc->fd, &cmd, sizeof (cmd));
}

static int sock_out_fragment (struct out_descr *desc, struct rect r, const struct pixel *px)
{
	struct sock_cmd cmd;

	memset (&cmd, 0, sizeof (cmd));
	cmd.sign = SIGN;
	cmd.cmd = _FRAG;
	cmd.data.frag.l = r.left;
	cmd.data.frag.t = r.top;
	cmd.data.frag.w = r.width;
	cmd.data.frag.h = r.height;
	write (desc->fd, &cmd, sizeof (cmd));
	write (desc->fd, px, r.width*r.height*sizeof(struct pixel));

	return 0;
}

static struct out_handler sock_handler = {
	stream_open:	sock_stream_open,
	stream_close:	sock_stream_close,
	frame_open:	sock_frame_open,
	frame_close:	sock_frame_close,
	out_fragment:	sock_out_fragment,
};

// output API

int out_open (struct out_descr *desc, const char *name, enum out_type type, int width, int height, int frames)
{
	desc->width = width;
	desc->height = height;
	desc->frame = 0;
	desc->frames = frames;
	desc->type = type;
	desc->name = name;

	desc->fd = -1;

	switch (type) {
	case out_type_tga:	desc->handler = &tga_handler;	break;
	case out_type_sock:	desc->handler = &sock_handler;	break;
	default:		desc->handler = NULL;		return -1;
	}

	if (desc->handler->stream_open (desc) < 0) return -1; 
	return desc->handler->frame_open (desc);
}

int out_fragment (struct out_descr *desc, struct rect r, const struct pixel *px)
{
	if (!desc) return -1;
	if (!desc->handler) return -1;
	return desc->handler->out_fragment(desc, r, px);
}

int out_close (struct out_descr *desc)
{
	int	res;
	
	if (!desc) return -1;
	if (!desc->handler) return -1;
	
	res = desc->handler->stream_close (desc);
	
	memset (desc, 0, sizeof (struct out_descr));
	desc->fd = -1;

	fprintf(stderr, "Output closed\n");
	
	return res;
}

int out_next (struct out_descr *desc)
{
	if (!desc) return -1;
	if (!desc->handler) return -1;
	
	desc->handler->frame_close (desc);
	desc->frame++;
	fprintf(stderr, "Next\n");
	if (desc->frame >= desc->frames) {
		return out_close (desc);
	} else {
		return desc->handler->frame_open (desc);
	}
}
