#define DMPI_IMPLEMENTATION
#include "dmpi.h"

#include <mpi.h>

#ifdef DEBUG
#define dprintf(x...) fprintf(stderr,x)
#else
#define dprintf(x...)
#endif

// type convertors 

// errcode
static inline int mpierrcode2errcode (int errcode)
{
  switch (errcode) {
  case MPI_SUCCESS: return _MPI_SUCCESS;
  case MPI_ERR_OTHER: return _MPI_ERR_OTHER;
  default: return _MPI_ERR_OTHER; // FIXME
  }
}

static inline int errcode2mpierrcode (int errcode)
{
  switch (errcode) {
  case _MPI_SUCCESS: return MPI_SUCCESS;
  case _MPI_ERR_OTHER: return MPI_ERR_OTHER;
  default: return MPI_ERR_OTHER; // FIXME
  }
}

// status
static inline MPI_Status* real_status (_MPI_Status *stat)
{
  //assert(sizeof(stat->realstatus) >= sizeof(MPI_Status)); // FIXME
  return ((MPI_Status*)&stat->real_status);
}

static inline void mpistatus2status(_MPI_Status *dst, const MPI_Status *src)
{
  dst->MPI_SOURCE = src->MPI_SOURCE;
  dst->MPI_TAG = src->MPI_TAG;
  dst->MPI_ERROR = src->MPI_ERROR;
}

static inline void status2mpistatus(MPI_Status *dst, const _MPI_Status *src)
{
  dst->MPI_SOURCE = src->MPI_SOURCE;
  dst->MPI_TAG = src->MPI_TAG;
  dst->MPI_ERROR = src->MPI_ERROR;
}

// datatype
static inline MPI_Datatype datatype2mpidatatype(_MPI_Datatype dt)
{
  switch (dt) {
  case _MPI_BYTE: return MPI_BYTE;
  default: return MPI_DATATYPE_NULL;
  }
}

// comm
static inline MPI_Comm comm2mpicomm(_MPI_Comm comm)
{
  switch (comm) {
  case _MPI_COMM_WORLD: return MPI_COMM_WORLD;
  case _MPI_COMM_SELF: return MPI_COMM_SELF;
  case _MPI_COMM_NULL: return MPI_COMM_NULL;
  default: return MPI_COMM_NULL; // FIXME assert
  }
}

// tag
static inline int tag2mpitag (int tag)
{
  return (tag == _MPI_ANY_TAG)?MPI_ANY_TAG:tag;
}

// nodeid
static inline int id2mpiid (int id)
{
  return (id == _MPI_ANY_SOURCE)?MPI_ANY_SOURCE:id;
}

// request
static inline MPI_Request* req2mpireq (_MPI_Request *req)
{
  //assert(sizeof(_MPI_Request) >= sizeof(MPI_Request)); // FIXME
  return (MPI_Request*) req;
}




// Function wrappers

int _MPI_Get_count (_MPI_Status *stat, _MPI_Datatype dtype, int *count) 
{
  MPI_Status *rstat;
  int rc;

  rstat = real_status(stat);
  status2mpistatus(rstat, stat);
  rc = MPI_Get_count(rstat, datatype2mpidatatype(dtype), count);
  mpistatus2status(stat, rstat);
  
  return mpierrcode2errcode(rc);
}

int _MPI_Iprobe (int src, int tag, _MPI_Comm comm, int *flag, _MPI_Status *stat)
{
  MPI_Status *rstat;
  int rc;

  rstat = real_status(stat);
  status2mpistatus(rstat, stat);
  rc = MPI_Iprobe (id2mpiid(src), tag2mpitag(tag), comm2mpicomm(comm), flag, rstat);
  mpistatus2status(stat, rstat);
  
  return mpierrcode2errcode(rc);
}

int _MPI_Recv(void *buf, int count, _MPI_Datatype dtype,
	      int src, int tag, _MPI_Comm comm, _MPI_Status *stat)
{
  MPI_Status *rstat;
  int rc;
  
  rstat = real_status(stat);
  status2mpistatus(rstat, stat);
  rc = MPI_Recv(buf, count, datatype2mpidatatype(dtype), id2mpiid(src), tag2mpitag(tag), comm2mpicomm(comm), rstat);
  mpistatus2status(stat, rstat);
  
  return mpierrcode2errcode(rc);
}

int _MPI_Send(void *buf, int count, _MPI_Datatype dtype,
	      int dest, int tag, _MPI_Comm comm)
{
  int rc = MPI_Send(buf, count, datatype2mpidatatype(dtype), id2mpiid(dest), tag2mpitag(tag), comm2mpicomm(comm));
  return mpierrcode2errcode(rc);
}

int _MPI_Isend(void *buf, int count, _MPI_Datatype dtype, int dest, 
	       int tag, _MPI_Comm comm, _MPI_Request *req) {
  
  int rc = MPI_Isend(buf, count, datatype2mpidatatype(dtype), id2mpiid(dest), tag2mpitag(tag),
		  comm2mpicomm(comm), req2mpireq(req));
  return mpierrcode2errcode(rc);
}

int _MPI_Test(_MPI_Request *req, int *flag, _MPI_Status *stat) {
  MPI_Status *rstat = MPI_STATUS_IGNORE;
  int rc;
  
  if (stat != _MPI_STATUS_IGNORE) {
    rstat = real_status(stat);
    status2mpistatus(rstat, stat);
  }
  
  dprintf("DMPI::TEST Request %p flag %p  _MPI status size %d MPI Status size %d\n",
		  (MPI_Request*) req, flag, sizeof(_MPI_Status), sizeof(MPI_Status));
  rc = MPI_Test(req2mpireq(req), flag, rstat);
  
  if (stat != _MPI_STATUS_IGNORE) {
    mpistatus2status(stat, rstat);
  }
  
  return mpierrcode2errcode(rc);
}

int _MPI_Wait(_MPI_Request *req, _MPI_Status *stat) {
  MPI_Status *rstat = MPI_STATUS_IGNORE;
  int rc;
  
  if (stat != _MPI_STATUS_IGNORE) {
    rstat = real_status(stat);
    status2mpistatus(rstat, stat);
  }
  
  dprintf("DMPI::WAIT Request %p _MPI status size %d MPI Status size %d\n",
		  (MPI_Request*) req, sizeof(_MPI_Status), sizeof(MPI_Status));
  rc = MPI_Wait(req2mpireq(req), rstat);
  
  if (stat != _MPI_STATUS_IGNORE) {
    mpistatus2status(stat, rstat);
  }
  
  return mpierrcode2errcode(rc);
}

int _MPI_Barrier (_MPI_Comm comm)
{
  return mpierrcode2errcode(MPI_Barrier(comm2mpicomm(comm)));
}

int _MPI_Comm_rank (_MPI_Comm comm, int *prank)
{
  return mpierrcode2errcode(MPI_Comm_rank(comm2mpicomm(comm), prank));
}

int _MPI_Comm_size (_MPI_Comm comm, int *psize)
{
  return mpierrcode2errcode(MPI_Comm_size(comm2mpicomm(comm), psize));
}

int _MPI_Finalize (void)
{
  return mpierrcode2errcode(MPI_Finalize());
}

double _MPI_Wtime (void)
{
  return MPI_Wtime();
}

int _MPI_Error_string (int errcode, char *msg, int *plen)
{
  return mpierrcode2errcode(MPI_Error_string(errcode2mpierrcode(errcode), msg, plen)); // FIXME conversion errors
}

#ifdef LAM_WANT_IMPI

static int _MPI_Init_check(int *c,char ***b)
{
  int res;
  int found;
  int *client_size;
  int *client_color;
  int *host_size;
  int *host_color;

  res = MPI_Init(c, b);
  if (res != MPI_SUCCESS) return _MPI_ERR_OTHER;

  MPI_Comm_get_attr (MPI_COMM_WORLD, IMPI_CLIENT_SIZE, &client_size, &found);
  if (res != MPI_SUCCESS || !found) return _MPI_ERR_OTHER;

  MPI_Comm_get_attr (MPI_COMM_WORLD, IMPI_CLIENT_COLOR, &client_color, &found);
  if (res != MPI_SUCCESS || !found) return _MPI_ERR_OTHER;

  MPI_Comm_get_attr (MPI_COMM_WORLD, IMPI_HOST_SIZE, &host_size, &found);
  if (res != MPI_SUCCESS || !found) return _MPI_ERR_OTHER;

  MPI_Comm_get_attr (MPI_COMM_WORLD, IMPI_HOST_COLOR, &host_color, &found);
  if (res != MPI_SUCCESS || !found) return _MPI_ERR_OTHER;

  dprintf ("IMPI: topology ids: c(%d/%d) h(%d/%d)\n",
    *client_color, *client_size, *host_color, *host_size);

  return _MPI_SUCCESS;
}

#else // LAM_WANT_IMPI

int _MPI_Init_check(int *c,char ***b)
{
  return (MPI_Init(c,b) == MPI_SUCCESS)?_MPI_SUCCESS:_MPI_ERR_OTHER;
}

#endif // LAM_WANT_IMPI

int _MPI_Hello(int *c,char ***b)
{
  int i;
  dprintf("DMPI Hello -xxx\n");
  dprintf("DMPI argc %p\n",c);
  dprintf("DMPI argc %d\n",*c);
  for (i=0; i<(*c); i++) {
    dprintf("DMPI argv[%d] %s \n", i, (*b)[i]);
  }
  dprintf("DMPI MPI_Init %p \n",&MPI_Init);
    
  dprintf("MPI_Init %x \n",((int*)&MPI_Init)[0]); 
  dprintf("MPI_Init %x \n",((int*)&MPI_Init)[1]); 

  int res = _MPI_Init_check(c,b);
  dprintf("DMPI Bye-bye\n");
  return res;
}

// entry points
struct mpi_portal_t global_mpi_portal = {
    DMPI_VERSION,
    DMPI_REVISION,
    MPI_NAME,
    BUILD_INFO,
#ifndef DEBUG
    _MPI_Init_check,
#else
    _MPI_Hello,
#endif 
    _MPI_Error_string,
    _MPI_Barrier,
    _MPI_Comm_rank,
    _MPI_Comm_size,
    _MPI_Finalize,
    _MPI_Get_count,
    _MPI_Iprobe,
    _MPI_Recv,
    _MPI_Send,
    _MPI_Isend,
    _MPI_Test,
    _MPI_Wait,
    _MPI_Wtime,
};
