#define DMPI_IMPLEMENTATION
#include "dmpi.h"
#include <assert.h>
#include <unistd.h> // for gethostname
#include <sys/stat.h>
#include <sys/time.h>

struct uni_msg_t {
  struct uni_msg_t *next;
  int tag;
  int count;
  unsigned char msg[0];
} *uni_msgs = NULL;


struct uni_msg_t **get_by_tag(int tag) 
{
  struct uni_msg_t **ppmsg = &uni_msgs;
  while (*ppmsg) {
    if (tag == _MPI_ANY_TAG || (*ppmsg)->tag == tag) return ppmsg;
    ppmsg = &(*ppmsg)->next;
  }
  return NULL;
}

static int uni_MPI_Init (int *pargc, char ***pargv) 
{
  return _MPI_SUCCESS;
}

static int uni_MPI_Error_string (int errcode, char *msg, int *plen)
{
  switch (errcode) {
  case _MPI_SUCCESS:
    strcpy (msg, "no error");
    *plen = 8;
    break;
  case _MPI_ERR_OTHER:
    strcpy (msg, "general MPI error");
    *plen = 17;
    break;
  default:
    *plen = sprintf (msg, "invalid MPI error code %d", errcode);
    break;
  }
  return _MPI_SUCCESS;
}

static int uni_MPI_Barrier (_MPI_Comm comm) 
{
  return _MPI_SUCCESS;
}

static int uni_MPI_Comm_rank (_MPI_Comm comm, int *rank) 
{
  *rank = 0;
  return _MPI_SUCCESS;
}

static int uni_MPI_Comm_size (_MPI_Comm comm, int *psize)
{
  *psize = 1;
  return _MPI_SUCCESS;
}

static int uni_MPI_Finalize (void) 
{
  return _MPI_SUCCESS;
}

static int uni_MPI_Get_count (_MPI_Status *stat, _MPI_Datatype dtype, int *count)
{
  struct uni_msg_t **ppmsg = get_by_tag(stat->MPI_TAG);
  if (ppmsg) {
    *count = (*ppmsg)->count;
  } else {
    *count = 0; // ??? 
  }
  return _MPI_SUCCESS;
}

static int uni_MPI_Iprobe (int src, int tag, _MPI_Comm comm, int *flag, 
		    _MPI_Status *stat)
{
  struct uni_msg_t **ppmsg = get_by_tag(tag);
  if (ppmsg) {
    *flag = 1;
    stat->MPI_TAG = (*ppmsg)->tag;
  } else {
    *flag = 0;
  }
  return _MPI_SUCCESS;
}

static int uni_MPI_Recv (void *buf, int count, _MPI_Datatype dtype,
		  int src, int tag, _MPI_Comm comm, _MPI_Status *stat)
{
  struct uni_msg_t **ppmsg = get_by_tag(tag), *ptmp;
  if (ppmsg) {
    int cpycnt = count < (*ppmsg)->count ? count : (*ppmsg)->count;
    memcpy(buf, (*ppmsg)->msg, cpycnt);
    ptmp = *ppmsg;
    *ppmsg = (*ppmsg)->next;
    DMPI_FREE(ptmp);
  }
  return _MPI_SUCCESS;
}
  
static int uni_MPI_Send (void *buf, int count, _MPI_Datatype dtype, int dest,
		  int tag, _MPI_Comm comm) 
{
  struct uni_msg_t *pmsg = (struct uni_msg_t*) 
    DMPI_MALLOC(sizeof(*pmsg) + count);
  assert(dest == 0);
  pmsg->next = uni_msgs;
  pmsg->tag = tag;
  pmsg->count = count;
  memcpy(pmsg->msg, buf, count);
  uni_msgs = pmsg;
  return _MPI_SUCCESS;
}

static int uni_MPI_Isend (void *buf, int count, _MPI_Datatype dtype, int dest,
		   int tag, _MPI_Comm comm, _MPI_Request* rq) 
{
  return uni_MPI_Send(buf, count, dtype, dest, tag, comm); // Just ignore `rq' here
}

static int uni_MPI_Test (_MPI_Request *req, int *flag, _MPI_Status *stat) {
  *flag = 1;   // Of course, already sent :-)
  assert(stat == (typeof(stat))_MPI_STATUS_IGNORE); // User! Keep program simple, really!
  return _MPI_SUCCESS;
}

static int uni_MPI_Wait (_MPI_Request *req, _MPI_Status *stat) {
  assert(stat == (typeof(stat))_MPI_STATUS_IGNORE); // User! Keep program simple, really!
  return _MPI_SUCCESS;
}

static double uni_MPI_Wtime () {
  struct timeval tv;
  assert(!gettimeofday(&tv,NULL));
  return tv.tv_sec+tv.tv_usec/1000000.;
}

static struct mpi_portal_t unicomputer_mpi_portal = {
  DMPI_VERSION,
  DMPI_REVISION,
  "unicomputer",
  "",
  uni_MPI_Init,
  uni_MPI_Error_string,
  uni_MPI_Barrier,
  uni_MPI_Comm_rank,
  uni_MPI_Comm_size,
  uni_MPI_Finalize,
  uni_MPI_Get_count,
  uni_MPI_Iprobe,
  uni_MPI_Recv,
  uni_MPI_Send,
  uni_MPI_Isend,
  uni_MPI_Test,
  uni_MPI_Wait,
  uni_MPI_Wtime,
};

struct mpi_portal_t *mpi_portal = &unicomputer_mpi_portal;



// See if we have been started with mpirun.  If so, load suitable 
// MPI run-time library.  If started without mpirun,
// use the built-in "unicomputer MPI" stub.

char dmpi_so[MAXSOPATH];
int  dmpi_meta;

static int dmpi_autodetect()
{
  struct stat statbuf;
  void *handle = NULL;
  const char *mpi = "default";
  int i;

  if (getenv("DMPI"))
    mpi = getenv("DMPI");
  else if (getenv("GLOBUS_DUROC_SUBJOB_INDEX"))
    mpi="mpich-g2"; 
  else if (getenv("LAMRANK"))
    mpi = "lam";
  else if (getenv("SCAMPI_PROCESS_PARAM"))
    mpi = "scali";
  else if (getenv("DMPI_DEFAULT"))
    mpi = getenv("DMPI_DEFAULT");

  // Detect metacluster
  dmpi_meta = !strcmp(mpi,"mpich-g2");

#ifdef DEBUG
  fprintf(stderr, "DMPI selection: %s\n", mpi);
#endif

  const char* sodir[] = { DMPI_LOCATION, getenv("HOME"), "" };

  for (i = 0; i < sizeof(sodir)/sizeof(sodir[0]); i++) {
    snprintf(dmpi_so, sizeof(dmpi_so)-1, "%s%slibdmpi_%s.so", sodir[i], *sodir[i]?"/":"", mpi);
    if (stat(dmpi_so,&statbuf)) continue;
    if (!(handle = dlopen(dmpi_so, RTLD_NOW)))  {
      fprintf(stderr, "%s\n", dlerror());
      return -1;
    } else
      break;
  }

  if (!handle) {
    if (strcmp(mpi,"default") != 0) {
      char hostname[0x100];
      memset(hostname, 0, sizeof(hostname));
      gethostname(hostname, sizeof(hostname));
      fprintf(stderr, "Unable load DMPI library trying:\n");
      for (i = 0; i < sizeof(sodir)/sizeof(sodir[0]); i++) 
	fprintf(stderr, "    %s%slibdmpi_%s.so\n", sodir[i], *sodir[i]?"/":"", mpi);
      fprintf(stderr,   " at host `%s'\n", hostname);
      return -1;
    } else {
#ifdef DEBUG
      fprintf(stderr, "Loading of libdmpi_%s.so failed, using unicomputer stub\n", mpi);
#endif
      snprintf(dmpi_so, sizeof(dmpi_so)-1, "%s", "unicomputer stub");
      return 0;
    }
  }
  
#ifdef DEBUG
  fprintf(stderr,"DMPI: before dlsym\n");
#endif 
  mpi_portal = (struct mpi_portal_t *) dlsym(handle, "global_mpi_portal");
#ifdef DEBUG
  fprintf(stderr,"DMPI: after dlsym\n");
#endif 
  if (!mpi_portal) {
    fprintf(stderr, "dlsym: %s\n", dlerror());
    return -1;
  }

  if (mpi_portal->version != DMPI_VERSION) {
    fprintf(stderr, "DMPI: invalid version %d (%d expected)\n", mpi_portal->version, DMPI_VERSION);
    return -1;
  }
//  fprintf (stderr, "DMPI Driver v%d r%s.\n", mpi_portal->version, mpi_portal->revision);
  return 0;
}

static int dmpi_initialize()
{
  static int initialized = 0;
  if (!initialized) {
     initialized = 1;
     return dmpi_autodetect();
  }
  return 0;
}	

int PMPI_Init (int *a, char ***b)
{  
  if (dmpi_initialize () < 0) {
    fprintf(stderr,"DMPI initialization failure\n");
    return _MPI_ERR_OTHER;
  }
#ifdef DEBUG
  fprintf(stderr, "DMPI: Calling _MPI_Init functions \n");
#endif 
  int res = mpi_portal->_MPI_Init(a,b);
#ifdef DEBUG 
  fprintf(stderr, "DMPI: After _MPI_Init functions \n");
#endif 
  return res ; 
}

//#pragma weak MPI_Init = PMPI_Init

int MPI_Init (int *a, char ***b)  __attribute__ ((weak));
int MPI_Init (int *a, char ***b) { return PMPI_Init(a,b); }
/*
int PMPI_Comm_rank (_MPI_Comm comm, int *rank)
{
  if (dmpi_initialize () < 0)
    return _MPI_ERR_OTHER;
  return mpi_portal->_MPI_Comm_rank(comm, rank);
}

//#pragma weak MPI_Comm_rank = PMPI_Comm_rank

int MPI_Comm_rank (_MPI_Comm comm, int *rank)  __attribute__ ((weak));
int MPI_Comm_rank (_MPI_Comm comm, int *rank) { return PMPI_Comm_rank(comm,rank); }
*/
