#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>
#include <dmpi.h>

struct mem_list {
  void *addr;
  long size;
  struct mem_list *prev;
  struct mem_list *next;
};

struct message_info {
  int count;
  MPI_Datatype type;
  int source;
  int dest;
  int tag;
  int len;
};

static int checkpoint_flag = 0;
static char checkpoint_filename[256];
static int checkpoint_interval;
static void *stack_bottom_addr;
static void *data_start_addr = 0;
static void *sbrk_addr = 0;
static struct mem_list *mem = 0;
jmp_buf env;

static void set_alarm();
int restart(char *filename);

/********************************************************************/

static void mem_add_chunk(void *addr, long size)
{
  struct mem_list *ptr1, *ptr2;
  
  if (!mem) {
    ptr1 = (struct mem_list *) malloc(sizeof(struct mem_list));
    ptr1->addr = addr;
    ptr1->size = size;
    ptr1->prev = 0;
    ptr1->next = 0;
    mem = ptr1;
  } else {
    for (ptr2 = mem; ptr2->next && ptr2->next->addr <= addr; ptr2 = ptr2->next)
      ;
    if (ptr2->addr > addr || ptr2->addr + ptr2->size < addr) {
      ptr1 = (struct mem_list *) malloc(sizeof(struct mem_list));
      ptr1->addr = addr;
      ptr1->size = size;
      if (ptr2->addr > addr) {
        ptr1->prev = 0;
        ptr1->next = ptr2;
        ptr2->prev = ptr1;
        mem = ptr1;
      } else {
        ptr1->prev = ptr2;
        ptr1->next = 0;
        ptr2->next = ptr1;
      }
      ptr2 = ptr1;
    } else
      if (ptr2->addr + ptr2->size < addr + size)
        ptr2->size = addr + size - ptr2->addr;
    
    while (ptr2->next && ptr2->next->addr + ptr2->next->size < ptr2->addr + ptr2->size) {
      ptr1 = ptr2->next;
      ptr2->next = ptr2->next->next;
      ptr2->next->prev = ptr2;
      free(ptr1);
    }        
    if (ptr2->next && ptr2->next->addr <= ptr2->addr + ptr2->size) {
      ptr2->size = ptr2->next->addr + ptr2->next->size - ptr2->addr;
      ptr1 = ptr2->next;
      ptr2->next = ptr2->next->next;
      ptr2->next->prev = ptr2;
      free(ptr1);
    }      
  }
}

/********************************************************************/

static void load_mpi()
{
  char filename[256];
  int fd;
  MPI_Request request;
  struct message_info mes;
  char *buf;

  sprintf(filename, "%s.mpi", checkpoint_filename);
  fd = open(filename, O_RDONLY, 0644);
  while (read(fd, &mes, sizeof(struct message_info)) > 0) {
    buf = malloc(mes.len);
    read(fd, buf, mes.len);
    MPI_Isend(buf, mes.count, mes.type, mes.dest, mes.tag, MPI_COMM_WORLD, &request);
    free(buf);
  }
  close(fd);
}

/********************************************************************/

static void save_mpi()
{
  char filename_tmp[256], filename[256];
  int rank, fd, flag, type_size, count;
  MPI_Status status;
  struct message_info mes;
  char *buf;

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  sprintf(filename_tmp, "%s.mpi.tmp", checkpoint_filename);
  sprintf(filename, "%s.mpi", checkpoint_filename);
  fd = open(filename_tmp, O_CREAT | O_TRUNC | O_WRONLY, 0644);

  flag = 1;
  while (flag) {
    MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &status);
    if (flag) {
//      MPI_Type_size(MPI_BYTE, &type_size);
      type_size = 1;
      MPI_Get_count(&status, MPI_BYTE, &count);
      buf = (char *) malloc(type_size * count);
      MPI_Recv(buf, count, MPI_BYTE, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
      mes.count = count;
      mes.type = MPI_BYTE;
      mes.source = status.MPI_SOURCE;
      mes.dest = rank;
      mes.tag = status.MPI_TAG;
      mes.len = type_size * count;
      write(fd, &mes, sizeof(struct message_info));
      write(fd, buf, mes.len);
      free(buf);
    }
  }
  fsync(fd);
  close(fd);
  rename(filename_tmp, filename);

  load_mpi();
}

/********************************************************************/

static void save_heap(int fd)
{
  struct mem_list *cur;

  for (cur = mem; cur; cur = cur->next) {
    write(fd, (void *) &cur->addr, sizeof(void *));
    write(fd, &cur->size, sizeof(long));
    write(fd, cur->addr, cur->size);
  }
}

/********************************************************************/

static void save_stack(int fd)
{
  void *stack_top_addr;
  long stack_size;
  int dummy;

  stack_top_addr = (void *) (&dummy - 100);
  stack_top_addr = (void *) ((unsigned long)stack_top_addr - ((unsigned long) stack_top_addr % getpagesize()));
    
  write(fd, (void *) &stack_top_addr, sizeof(void *));
  stack_size = stack_bottom_addr - stack_top_addr;
  write(fd, &stack_size, sizeof(long));
  write(fd, stack_top_addr, stack_size);
}

/********************************************************************/

static void checkpoint_handler()
{
  struct mem_list *cur;
  int heap_chunks_count;
  char filename[256];
  int mpi_flag;
  int fd;

  alarm(0);

//  MPI_Initialized(&mpi_flag);
  for (cur = mem, heap_chunks_count = 0; cur; cur = cur->next, heap_chunks_count ++)
    ;

  if (!setjmp(env)) {
    sprintf(filename, "%s.tmp", checkpoint_filename);
    fd = open(filename, O_CREAT | O_TRUNC | O_WRONLY, 0644);

    write(fd, (void *) &sbrk_addr, sizeof(void *));
    write(fd, &heap_chunks_count, sizeof(int));
    save_heap(fd);

    save_stack(fd);

    fsync(fd);
    close(fd);
    rename(filename, checkpoint_filename);

//    if (mpi_flag)
      save_mpi();
      
  } else {
//    if (mpi_flag)
      load_mpi();
    return;
  }
//  set_alarm();
}

/********************************************************************/

static void set_alarm()
{
  struct sigaction act;

  sigemptyset(&act.sa_mask);
  act.sa_flags = SA_RESTART;
  act.sa_handler = checkpoint_handler;
  sigaction(SIGALRM, &act, 0);
  alarm(checkpoint_interval);
}

/********************************************************************/

extern int etext;
static void *ptr;

static void segv_handler(int sig, siginfo_t *info, void *v)
{
  if(!data_start_addr)
    ptr += 1024;
  else
    stack_bottom_addr = (void *) (info->si_addr - 1);
  siglongjmp(env, 1);
}

/********************************************************************/

int checkpoint(char *filename, int interval)
{
  int dummy;
  struct sigaction act;
  int page_size;

  checkpoint_flag = 1;
  strcpy(checkpoint_filename, filename);
  checkpoint_interval = interval;

  page_size = getpagesize();

  ptr = (void *) &etext;
  ptr = (void *) (((unsigned long) ptr / page_size + 1) * page_size);
  sigemptyset(&act.sa_mask);
  act.sa_flags = SA_SIGINFO;
  act.sa_handler = (__sighandler_t) segv_handler;
  sigaction(SIGSEGV, &act, 0);
  sigsetjmp(env, 1);
  *((char *) ptr) = *((char *) ptr);
  data_start_addr = ptr;

  stack_bottom_addr = 0;
  sigemptyset(&act.sa_mask);
  act.sa_flags = SA_SIGINFO | SA_RESETHAND;
  act.sa_handler = (__sighandler_t) segv_handler;
  sigaction(SIGSEGV, &act, 0);
  ptr = (void *) &dummy;
  sigsetjmp(env, 1);
  while (!stack_bottom_addr) {
    *((char *) ptr) = *((char *) ptr);
    ptr ++;
  }

  sbrk_addr = sbrk(0);
  mem_add_chunk(data_start_addr, sbrk_addr - data_start_addr);

  set_alarm();
  return 0;
}

/********************************************************************/

int checkpoint_init(int argc, char **argv)
{
  int checkpoint_flag = 0, interval = 0, restart_flag = 0;
  char checkpoint_filename[256], restart_filename[256];
  int mpi_flag = 0, rank;
  int i;

  for (i = 1; i < argc; i ++)
  {
    if (!strcmp(argv[i], "-ckpt-checkpoint")) {
      i ++;
      if (i < argc) {
        strcpy(checkpoint_filename, argv[i]);
        checkpoint_flag = 1;
      } else {
        fprintf(stderr, "checkpoint is disabled - please enter checkpoint file name\n");
        return -1;
      }
    } else if (!strcmp(argv[i], "-ckpt-interval")) {
      i ++;
      if (i < argc)
        interval = atoi(argv[i]);
      else {
        fprintf(stderr, "checkpoint is disabled - please enter checkpoint time interval\n");
        return -1;
      }
    } else if (!strcmp(argv[i], "-ckpt-restart")) {
      i ++;
      if (i < argc) {
        strcpy(restart_filename, argv[i]);
        restart_flag = 1;
      } else {
        fprintf(stderr, "restart is disabled - please enter restart file name\n");
        return -1;
      }
    }
  }

  if (checkpoint_flag && interval <= 0) {
    checkpoint_flag = 0;
    fprintf(stderr, "checkpoint is disabled - please enter/correct checkpoint time interval\n");
    return -1;
  }

  if (restart_flag || checkpoint_flag) {
//    MPI_Initialized(&mpi_flag);
//    if (mpi_flag) {
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
      if (restart_flag)
        sprintf(restart_filename, "%s.%d", restart_filename, rank);
      if (checkpoint_flag)
        sprintf(checkpoint_filename, "%s.%d", checkpoint_filename, rank);
//    }

    if (restart_flag) {
      if (access(restart_filename, R_OK)) {
//        if (mpi_flag)
          MPI_Finalize();
        exit(0);
      }
      if (restart(restart_filename) == -1)
        return -2;
//      else
//        fprintf(stderr, "restart is completed\n");
    }

    if (checkpoint_flag) {
      if (checkpoint(checkpoint_filename, interval) == -1)
        return -1;
//      else
//        fprintf(stderr, "checkpoint is completed\n");
    }
  }
  return 0;
}

/********************************************************************/

void *ckptsbrk(int incr)
{
  void *prev_sbrk_addr;
  void *ret;

  if(!checkpoint_flag || !incr)
    return sbrk(incr);

  ret = sbrk(incr);
  if ((int) ret == -1)
    return (void *) -1;

  if(sbrk_addr < sbrk(0)) {
    prev_sbrk_addr = sbrk_addr;
    sbrk_addr = sbrk(0);
    if (mem)
      mem_add_chunk(prev_sbrk_addr, sbrk_addr - prev_sbrk_addr);
  }
  return ret;
}
